//
// TabWidth: 4
// Compiler: WinAVR-20070525
//
// AquaCtrl Firmware for Aquarium Controller
// Copyright (C) 2008, 2009  Christian Roetzer
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#define VERSION     "0.14"

#define F_CPU 		20000000UL

#include <stdlib.h>
#include <string.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <util/delay.h>

#include "FreeRTOS.h"
#include "task.h"
#include "queue.h"
#include "semphr.h"
#include "croutine.h"

// ******************************************************************
// ****** Basic Data Types ******************************************
// ******************************************************************

typedef unsigned char byte;
typedef unsigned int  word;
typedef unsigned char bool;
#define true		  1
#define false         0

// A generic function pointer W/O parameters
typedef void (*VoidFuncPtrVoid)(void);
// A generic function pointer W/ a pointer to a parameter
typedef void (*VoidFuncPtrVoidPtr)(void *);
// A generic function pointer with a byte as argument
typedef void (*VoidFuncPtrByte)(byte);

// ******************************************************************
// ****** Serial Parameters *****************************************
// ******************************************************************

#define BAUDRATE0		115200
// Define the baudrate register value incl. rounding,
// double speed mode, U2X assumed to be 1
#define DIV(a,b)		(((a)+(b)/2L)/(b))
#define BAUDVAL0		(DIV(F_CPU/8,BAUDRATE0)-1)

// ******************************************************************
// ****** Switch Channel Bits ***************************************
// ******************************************************************

#define MSK_DAYLIGHT   	0x01
#define MSK_MOONLIGHT 	0x02
#define MSK_CO2        	0x04
#define MSK_HEATER     	0x08
#define MSK_PUMP       	0x10

// ******************************************************************
// ****** Hardware Abstraction Macros *******************************
// ******************************************************************

#define DCF77_IS_OFF	( (PIND & 0x04) == 0 )
#define DCF77_IS_ON		( (PIND & 0x04) != 0 )

#define DBG1_ON			{ PORTD |=  0x08; }
#define DBG1_OFF		{ PORTD &= ~0x08; }

#define BEEP_ON     	{ TCCR0A |=  (1 << COM0B0); }
#define BEEP_OFF    	{ TCCR0A &= ~(1 << COM0B0); }

#define DAYLIGHT_ON  	{ PORTB |=  0x01;           OnOffStates |=  MSK_DAYLIGHT;  }
#define DAYLIGHT_OFF 	{ PORTB &= ~0x01;           OnOffStates &= ~MSK_DAYLIGHT;  }
#define MOONLIGHT_ON	{ TCCR2A |=  (1 << COM2B1); OnOffStates |=  MSK_MOONLIGHT; }
#define MOONLIGHT_OFF	{ TCCR2A &= ~(1 << COM2B1); OnOffStates &= ~MSK_MOONLIGHT; }
#define CO2_ON       	{ PORTB |=  0x02;           OnOffStates |=  MSK_CO2;       }
#define CO2_OFF      	{ PORTB &= ~0x02;           OnOffStates &= ~MSK_CO2;       }
#define HEATER_ON    	{ PORTB |=  0x04;           OnOffStates |=  MSK_HEATER;    }
#define HEATER_OFF   	{ PORTB &= ~0x04;           OnOffStates &= ~MSK_HEATER;    }
#define PUMP_ON      	{ PORTB |=  0x08;           OnOffStates |=  MSK_PUMP;      }
#define PUMP_OFF     	{ PORTB &= ~0x08;           OnOffStates &= ~MSK_PUMP;      }

#define DAYLIGHT_IS_ON 	( OnOffStates & MSK_DAYLIGHT  )
#define MOONLIGHT_IS_ON	( OnOffStates & MSK_MOONLIGHT )
#define CO2_IS_ON      	( OnOffStates & MSK_CO2       )
#define HEATER_IS_ON   	( OnOffStates & MSK_HEATER    )
#define PUMP_IS_ON     	( OnOffStates & MSK_PUMP      )

#define LCD_E_HI	{ PORTA |=  0x04; }
#define LCD_E_LO	{ PORTA &= ~0x04; }
#define LCD_RS_1	{ PORTA |=  0x08; }
#define LCD_RS_0	{ PORTA &= ~0x08; }
#define LCD_WRITE(x)							\
	{											\
		PORTA = (PORTA & 0x0F) | ((x) & 0xF0);	\
		LCD_E_HI;								\
		_delay_loop_1(10);						\
		LCD_E_LO;								\
		_delay_loop_1(10);						\
	}
#define LCD_CMD(x)								\
	{											\
		LCD_RS_0;								\
		LCD_WRITE(x);							\
		LCD_WRITE((x) << 4);					\
	}
#define LCD_DATA(x)								\
	{											\
		LCD_RS_1;								\
		LCD_WRITE(x);							\
		LCD_WRITE((x) << 4);					\
	}

#define LCD_SET_CONTRAST(x)						\
	{											\
		portENTER_CRITICAL();					\
		{										\
			OCR1B = 300+((x)<<6);				\
		}										\
		portEXIT_CRITICAL();					\
	}

// ******************************************************************
// ****** Event Queue Codes *****************************************
// ******************************************************************

#define KEY_CODE		0x0F

#define KEY_ESC			0x01
#define KEY_DOWN		0x02
#define KEY_UP			0x03
#define KEY_ENTER		0x04

#define KEY_DAYLIGHT	0x09
#define KEY_MOONLIGHT	0x0A
#define KEY_CO2			0x0B
#define KEY_HEATER		0x0C
#define KEY_PUMP		0x08

#define SWITCHKEY_MIN	0x08
#define SWITCHKEY_MAX	0x0C

#define KEY_ACTION		0x30

#define KEY_PRESS		0x00
#define KEY_REPEAT		0x10
#define KEY_RELEASE		0x20

#define EV_SECOND		0x80
#define EV_SWITCH		0x81
#define EV_PAINT		0x82

// ******************************************************************
// ****** Date and Time *********************************************
// ******************************************************************

typedef word TTime;		// 0..MAXTIME
#define MAXTIME         1439
#define UNKNOWN_TIME    (MAXTIME+1)
#define TIME(HH,MM)		((HH)*60+(MM))
#define MINUTE(Time)	((Time)%60)
#define HOUR(Time)		((Time)/60)

// Convert a DCF byte from BCD to binary
#define BCD2BIN(x)      ((x & 0x0F) + 10*(x >> 4))

#define UNKNOWN_DAY		0

typedef struct
{
	byte Day;	 		// 1..31
	byte Month;	 		// 1..12
	byte Year;	 		// 0..99
} TDate;

static const byte DaysPerMonth[13] PROGMEM = { 0,31,29,31,30,31,30,31,31,30,31,30,31 };

#define INC_DATE(Date)					\
	{									\
		if ( (Date.Day == pgm_read_byte(&DaysPerMonth[Date.Month])) ||				\
			(((Date.Year & 3) != 0) && (Date.Month == 2) && (Date.Day == 28)) )		\
		{								\
			Date.Day = 1;				\
			if (Date.Month == 12)		\
			{							\
				Date.Month = 1;			\
				if (Date.Year == 99)	\
					Date.Year = 0;		\
				else					\
					Date.Year++;		\
			}							\
			else						\
				Date.Month++;			\
		}								\
		else							\
			Date.Day++;					\
	}

// ******************************************************************
// ****** Beep Types ************************************************
// ******************************************************************

#define BEEP_INFO		1
#define BEEP_WARNING	2
#define BEEP_ERROR		3

// ******************************************************************
// ****** Operating Modes *******************************************
// ******************************************************************

#define OPMODE_AUTO		0
#define OPMODE_MANUAL	1

// ******************************************************************
// ****** Dim/Switch Event Declarations *****************************
// ******************************************************************

#define MAXLISTITEMS	8

struct TDimEvent
{
	TTime Time;			
	byte DimSpeed;	// 0..99
	byte Power;		// 0..100
};
struct TDimEventList
{
	byte Count;
	struct TDimEvent Items[MAXLISTITEMS];
};
struct TSwitchEvent
{
	TTime Time;
	bool State;    	// false, true
};
struct TSwitchEventList
{
	byte Count;
	struct TSwitchEvent Items[MAXLISTITEMS];
};

// ******************************************************************
// ****** Configuration Variables ***********************************
// ******************************************************************

struct TCalPoint
{
	word Uncal;
	word Cal;
};

struct TCalPoints
{
	struct TCalPoint A;
	struct TCalPoint B;
};

struct TTemp
{
	bool Regulate; 		// Temp -> Heater
	word Nominal;		
	word Hysterese;
	word AlarmMin;
	word AlarmMax;
	struct TCalPoints CalPoints;
};

struct TpH				
{
	bool Regulate;		// CO2 -> pH
	word Nominal;		
	word Hysterese;
	word AlarmMin;
	word AlarmMax;
	struct TCalPoints CalPoints;
};

struct TConfig
{
	struct TTemp Temp;						//  17
	struct TpH pH;							//  17
	struct TDimEventList DaylightEvents;	//  33
	struct TDimEventList MoonlightEvents;	//  33
	struct TSwitchEventList CO2Events;		//  25
	struct TSwitchEventList HeaterEvents;	//  25
	struct TSwitchEventList PumpEvents;		//  25
	byte LCDContrast;						//   1
	// The magic word is the last entry. Therefore, if the
	// EEPROM is written the first time, it will get valid
	// AFTER all other values have been written.	
	word Magic;								//   2
											// ---
											// 178
};

// ******************************************************************
// ****** Channel Declarations **************************************
// ******************************************************************

struct TDimChannel
{
	byte CurrentPower;
	byte FinalPower;
	byte DiffPower;
	unsigned long DimSpeed;
	unsigned long BresAccu;
	
	struct TDimEventList *Events;
	VoidFuncPtrByte SetPower;
};

struct TSwitchChannel
{
	bool *Regulated;
	struct TSwitchEventList *Events;
	VoidFuncPtrByte Switch;	
};

// ******************************************************************
// ****** Prototypes ************************************************
// ******************************************************************

bool ReadConfig(void);
void DefaultConfig(void);
void Beep(byte BeepType);

void SwitchCO2(byte State);
void SwitchHeater(byte State);
void SwitchPump(byte State);

void SetDaylightPower(byte Power);
void SetMoonlightPower(byte Power);

// ******************************************************************
// ****** Global Variables ******************************************
// ******************************************************************

static struct TConfig Config;

static xQueueHandle ADCQueue;
static xQueueHandle DCFQueue;
static xQueueHandle EventQueue;
static xQueueHandle BeepQueue;
static xQueueHandle RXQueue;
static xQueueHandle TXQueue;

static word SensorTempADCUncal;
static word	SensorTempUncal;
static word	SensorTempCal;
static word SensorpHADCUncal;
static word SensorpHUncal;
static word SensorpHCal;

static byte OnOffStates;

static byte SystemTimeHundredth;	 	// 0..99
static byte SystemTimeSecond;	 		// 0..59
static TTime SystemTime;
static TDate SystemDate;

static word SystemTicks;
static int CrystalCal;
static TTime DCFTime;
static TDate DCFDate;

static byte OpMode;

#define CHANNEL_DAYLIGHT	0
#define CHANNEL_MOONLIGHT	1
static struct TDimChannel DimChannels[2] =
{
	{ 0, 0, 0, 0, 0, &Config.DaylightEvents , SetDaylightPower  },
	{ 0, 0, 0, 0, 0, &Config.MoonlightEvents, SetMoonlightPower }
};

#define CHANNEL_CO2			0
#define CHANNEL_HEATER		1
#define CHANNEL_PUMP		2
static struct TSwitchChannel SwitchChannels[3] =
{
	{ &Config.pH.Regulate  , &Config.CO2Events   , SwitchCO2    },
	{ &Config.Temp.Regulate, &Config.HeaterEvents, SwitchHeater },
	{  NULL                , &Config.PumpEvents  , SwitchPump   }
};

void InitGlobals(void)
{
	if (!ReadConfig())
		DefaultConfig();

	// A queue, which collects ADC values in exact
	// hundredth of seconds. Therefore, it also
	// is the base for the system time
	ADCQueue = xQueueCreate(10,sizeof(word));
	// A queue, which collects the bits from the DCF77 receiver
	DCFQueue = xQueueCreate(10,sizeof(byte));
	// A queue, which holds 10 bytes
    EventQueue = xQueueCreate(10,sizeof(byte));
	// A queue, which holds 10 bytes
    BeepQueue = xQueueCreate(10,sizeof(byte));
    // The serial receiption queue
    RXQueue = xQueueCreate(100,sizeof(char));
    // The serial transmission queue
    TXQueue = xQueueCreate(100,sizeof(char));

	SensorTempADCUncal = 512;	
	SensorTempUncal    = 2500;
	SensorTempCal      = 2500;
	SensorpHADCUncal   = 512;	
	SensorpHUncal      = 700;
	SensorpHCal        = 700;
	
	SystemDate.Day      = UNKNOWN_DAY;
	SystemDate.Month    = 1;
	SystemDate.Year     = 0;

	SystemTime          = UNKNOWN_TIME;
	SystemTimeSecond    = 0;
	SystemTimeHundredth = 0;

	OnOffStates = 0x00;		// All off

    SystemTicks = 0;
    CrystalCal = 0;			// typ. 1130 = 113.0ppm
	DCFTime = UNKNOWN_TIME;
	DCFDate.Day = UNKNOWN_DAY;
	DCFDate.Month = 0;
	DCFDate.Year = 0;
	
	OpMode = OPMODE_AUTO;
}

// ******************************************************************
// ****** Helper Functions ******************************************
// ******************************************************************

void WordToString(word Word,char *String,int Radix,int Width,bool FillZero)
{
	String += Width;
	*String-- = 0;
	int d;
	for (d=0; d<Width; d++)
	{
		if (!FillZero && Word==0 && d>0)
			*String-- = ' ';
		else
			*String-- = '0' + Word % Radix;
		Word /= Radix;
	}
}

void IntToString(int Int,char *String,int Radix,int Width)
{
	String += Width;
	*String-- = 0;
	bool Neg = Int < 0;
	if (Neg)
		Width--;
	else
		// Always convert to a negative(!) numbers in order to
		// get a correct result for -32768, which can't be
		// displayed as a positive number.
		Int = -Int;
	for ( ; Width>0; Width--)
	{
		char Char = Int % Radix;
		
		if (Char <= 9)
			*String = '0'-Char;
		else
			*String = 'A'-10-Char;
		String--;
		Int /= Radix;
		if (Int == 0)
			break;
	}
	if (Neg)
		*String-- = '-';
	for ( ; Width>0; Width--)
		*String-- = ' ';
}

// ******************************************************************
// ****** EEPROM Functions ******************************************
// ******************************************************************

#define EE_MAGIC		0x3A96

bool ReadConfig(void)
{
	byte *p;
	byte i;
	for (i=0, p=((byte *)&Config); i<sizeof(Config); i++, p++)
		*p = eeprom_read_byte((byte *)(word)i);
	if (Config.Magic != EE_MAGIC)
		return false;
	return true;
}

void WriteConfig(void)
{
	bool BytesWritten = false;
	byte *p;
	byte i;
	for (i=0, p=((byte *)&Config); i<sizeof(Config); i++, p++)
	{
		byte Value = eeprom_read_byte((byte *)(word)i);
		if (*p != Value)
		{
			// eeprom_write_byte disable interrupts, where necessary
			eeprom_write_byte((byte *)(word)i,*p);
			BytesWritten = true;
		}
	}
	if (BytesWritten)
		Beep(BEEP_INFO);
}

static void vTaskEEPROM(void *pvParameters)
{
	while (true)
	{
		WriteConfig();
		vTaskDelay(2000);
	}
}

void DefaultConfig(void)
{
	Config.Temp.Regulate = false;
	Config.Temp.Nominal = 2500;
	Config.Temp.Hysterese = 500;
	Config.Temp.AlarmMin = 2300;
	Config.Temp.AlarmMax = 2700;
	Config.Temp.CalPoints.A.Uncal = 512;
	Config.Temp.CalPoints.A.Cal = 512;
	Config.Temp.CalPoints.B.Uncal = 768;
	Config.Temp.CalPoints.B.Cal = 768;

	Config.pH.Regulate = false;
	Config.pH.Nominal = 700;
	Config.pH.Hysterese = 20;
	Config.pH.AlarmMin = 650;
	Config.pH.AlarmMax = 750;
	Config.pH.CalPoints.A.Uncal = 768;
	Config.pH.CalPoints.A.Cal = 768;
	Config.pH.CalPoints.B.Uncal = 512;
	Config.pH.CalPoints.B.Cal = 512;

	Config.DaylightEvents.Count = 0;
	Config.MoonlightEvents.Count = 0;
	Config.CO2Events.Count = 0;
	Config.HeaterEvents.Count = 0;
	Config.PumpEvents.Count = 0;

	Config.LCDContrast = 5;

	Config.Magic = EE_MAGIC;
}

// ******************************************************************
// ****** RS232 Functions *******************************************
// ******************************************************************

ISR( USART0_RX_vect )
{
	char Char = UDR0;
	if (xQueueSendFromISR(RXQueue,&Char,pdFALSE))
		taskYIELD();
}

ISR( USART0_UDRE_vect )
{
	char Char;
	signed portBASE_TYPE TaskWoken = pdFALSE;
	if (xQueueReceiveFromISR(TXQueue,&Char,&TaskWoken))
		UDR0 = Char;
	else
		UCSR0B &= ~(1 << UDRIE0);
}

int SerialGetChar(void)
{
	char Char;
	if(!xQueueReceive(RXQueue,&Char,0))
		return -1;
	return Char;
}

void SerialPutChar(char Char)
{
	if (xQueueSend(TXQueue,&Char,0) != pdPASS)
		return;
	portENTER_CRITICAL();
	{
		UCSR0B |= 1 << UDRIE0;
	}
	portEXIT_CRITICAL();
	
}

void SerialPutBCD(byte BCD)
{
	SerialPutChar('0'+(BCD>>4));
	SerialPutChar('0'+(BCD&0x0F));
}

void SerialPutBin(byte Bin)
{
	SerialPutChar('0'+(Bin/10));
	SerialPutChar('0'+(Bin%10));
}

void SerialPutString(const char *String)
{
	while (*String)
		SerialPutChar(*String++);
}

void SerialPutInt(int Int)
{
	static char String[1+5+1];
	IntToString(Int,String,10,6);
	SerialPutString(String);
}

#define DEBUGDCF77CHAR(x)	SerialPutChar(x)
#define	DEBUGDCF77BCD(x)	SerialPutBCD(x)
#define	DEBUGDCF77BIN(x)	SerialPutBin(x)
#define DEBUGDCF77INT(x)	SerialPutInt(x)

// ******************************************************************
// ****** Basic Switch and Dim Functions ****************************
// ******************************************************************

// Notes on the gamma table:
// - Index 0 is not used, because this is the 'Off'-state
// - Index 1 is the 0%-value
// - Index 100 is the 100%-value

const byte Gamma[101] PROGMEM =
{
	  0,  
      0,   1,   2,   3,   4,   5,   6,   7,   8,   9,
     10,  11,  12,  14,  15,  16,  17,  18,  20,  21,
     22,  24,  25,  26,  28,  29,  31,  32,  34,  35,
     37,  38,  40,  42,  43,  45,  47,  49,  50,  52, 
     54,  56,  58,  60,  62,  64,  66,  68,  71,  73,
     75,  77,  80,  82,  85,  87,  90,  92,  95,  97, 
    100, 103, 106, 108, 111, 114, 117, 120, 124, 127, 
    130, 133, 137, 140, 144, 147, 151, 154, 158, 162, 
    166, 170, 174, 178, 182, 187, 191, 195, 200, 204,
    209, 214, 219, 224, 229, 234, 239, 244, 250, 255
};

void SetDaylightPower(byte Power)
{
	if (Power > 0)
	{
		DAYLIGHT_ON
		OCR2A = 255-pgm_read_byte(&Gamma[Power]);
	}
	else
	{
		DAYLIGHT_OFF
		OCR2A = 255;
	}
}

void SetMoonlightPower(byte Power)
{
	if (Power > 0)
	{
		MOONLIGHT_ON
		OCR2B = pgm_read_byte(&Gamma[Power]);
	}
	else
	{
		MOONLIGHT_OFF
		OCR2B = 0;
	}
}

void SwitchCO2(byte State)
{
	if (State)
		CO2_ON
	else
		CO2_OFF
}

void SwitchHeater(byte State)
{
	if (State)
		HEATER_ON
	else
		HEATER_OFF
}

void SwitchPump(byte State)
{
	if (State)
		PUMP_ON
	else
		PUMP_OFF
}

// ******************************************************************
// ****** Beep Functions ********************************************
// ******************************************************************

void Beep(byte BeepType)
{
	xQueueSend(BeepQueue,&BeepType,0);
}

static void vTaskBeep(void *pvParameters)
{
	while (true)
	{
		byte BeepType;
		while (!xQueueReceive(BeepQueue,&BeepType,1))
			taskYIELD();
		switch (BeepType)
		{
		case BEEP_ERROR:
			BEEP_ON
			vTaskDelay(50);
			BEEP_OFF
			vTaskDelay(50);
		case BEEP_WARNING:
			BEEP_ON
			vTaskDelay(50);
			BEEP_OFF
			vTaskDelay(50);
		case BEEP_INFO:
			BEEP_ON
			vTaskDelay(50);
			BEEP_OFF
			vTaskDelay(50);
		}
	}
}

// ******************************************************************
// ****** LCD Functions *********************************************
// ******************************************************************

/*
	 -> 0xE1
	 -> 0xEF
	 -> 0xF5
	 -> 0xE2
	 -> [0]
	 -> [1]
	 -> [2]
	\ -> [3]
	 -> [4]
	> -> [5]
*/

const byte LCDCharPattern[] PROGMEM =
{
	// 
	0b00010001,
	0b00000100,
	0b00001010,
	0b00010001,
	0b00011111,
	0b00010001,
	0b00010001,
	0b00000000,
	// 
	0b00010001,
	0b00001110,
	0b00010001,
	0b00010001,
	0b00010001,
	0b00010001,
	0b00001110,
	0b00000000,
	// 
	0b00010001,
	0b00000000,
	0b00010001,
	0b00010001,
	0b00010001,
	0b00010001,
	0b00001110,
	0b00000000,
	// Backslash
	0b00000000,
	0b00010000,
	0b00001000,
	0b00000100,
	0b00000010,
	0b00000001,
	0b00000000,
	0b00000000,
	// 
	0b00001100,
	0b00010010,
	0b00010010,
	0b00001100,
	0b00000000,
	0b00000000,
	0b00000000,
	0b00000000,
	// ->
	0b00000000,
	0b00000100,
	0b00000110,
	0b00011111,
	0b00000110,
	0b00000100,
	0b00000000,
	0b00000000
};

#define ARROW_RIGHT		0x05

const char LCDCharTab[256] PROGMEM =
{
	// '?'  '?'  '?'  '?'  '?'  '?'  '?'  '?'  '?'  '?'  '?'  '?'  '?'  '?'  '?'  '?'
 	   '?', '?', '?', '?', '?',0x05, '?', '?', '?', '?', '?', '?', '?', '?', '?', '?',
	// '?'  '?'  '?'  '?'  '?'  '?'  '?'  '?'  '?'  '?'  '?'  '?'  '?'  '?'  '?'  '?'
 	   '?', '?', '?', '?', '?', '?', '?', '?', '?', '?', '?', '?', '?', '?', '?', '?',
	// ' '  '!'  '"'  '#'  '$'  '%'  '&'  '''  '('  ')'  '*'  '+'  ','  '-'  '.'  '/'
       ' ', '!', '"', '#', '$', '%', '&','\'', '(', ')', '*', '+', ',', '-', '.', '/',
	// '0'  '1'  '2'  '3'  '4'  '5'  '6'  '7'  '8'  '9'  ':'  ';'  '<'  '='  '>'  '?'
       '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', ':', ';', '<', '=', '>', '?',

	// '@'  'A'  'B'  'C'  'D'  'E'  'F'  'G'  'H'  'I'  'J'  'K'  'L'  'M'  'N'  'O'
       '@', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O',
	// 'P'  'Q'  'R'  'S'  'T'  'U'  'V'  'W'  'X'  'Y'  'Z'  '['  '\'  ']'  '^'  '_'
       'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', '[',0x03, ']', '^', '_',
	// '`'  'a'  'b'  'c'  'd'  'e'  'f'  'g'  'h'  'i'  'j'  'k'  'l'  'm'  'n'  'o'
       '`', 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o',
	// 'p'  'q'  'r'  's'  't'  'u'  'v'  'w'  'x'  'y'  'z'  '{'  '|'  '}'  '~'  ''
       'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', '{', '|', '}', '?', '?',

	// ''  ''  ''  ''  ''  ''  ''  ''  ''  ''  ''  ''  ''  ''  ''  ''
       '?', '?', ',', '?', '?', '?', '?', '?', '?', '?', '?', '?', 'O', '?', '?', '?',
	// ''  ''  ''  ''  ''  ''  ''  ''  ''  ''  ''  ''  'a'  ''  ''  ''
       '?', '?', '?', '?', '?', '?', '-', '-', '?', '?', '?', '?', '?', '?', 'z', 'Y',
	// ''  ''  ''  ''  ''  ''  ''  ''  ''  'c'  ''  ''  ''  ''  ''  ''
       '?', '?', '?', '?', '?', '?', '|', '?', '?', '?', '?', '<', '-', '-', '?', '?',
	// ''  ''  ''  ''  ''  ''  ''  ''  ''  ''  ''  ''  ''  ''  ''  ''
      0x04, '?', '?', '?','\'', '?', '?', '?', '?', '?',0x04, '>', '?', '?', '?', '?',

	// ''  ''  ''  ''  ''  ''  ''  ''  ''  ''  ''  ''  ''  ''  ''  ''
       'A', 'A', 'A', 'A',0x00, 'A', 'A', 'C', 'E', 'E', 'E', 'E', 'I', 'I', 'I', 'I',
	// ''  ''  ''  ''  ''  ''  ''  ''  ''  ''  ''  ''  ''  ''  ''  ''
       'D', 'N', 'O', 'O', 'O', 'O',0x01, 'x', 'O', 'U', 'U', 'U',0x02, 'Y', 'p',0xE2,
	// ''  ''  ''  ''  ''  ''  ''  ''  ''  ''  ''  ''  ''  ''  ''  ''
       'a', 'a', 'a', 'a',0xE1, 'a', '?', 'c', 'e', 'e', 'e', 'e', 'i', 'i', 'i', 'i',
	// ''  ''  ''  ''  ''  ''  ''  ''  ''  ''  ''  ''  ''  ''  ''  ''
       'o', 'n', 'o', 'o', 'o', 'o',0xEF, '-', 'o', 'u', 'u', 'u',0xF5, 'y', 'p', 'y'
};

void LCDInit(void)
{
	LCD_RS_0;
	_delay_ms(15);
	LCD_WRITE(0x30);
	_delay_ms(5);
	LCD_WRITE(0x30);
	_delay_ms(20);
	LCD_WRITE(0x30);
	_delay_ms(1);
	// Set 4-bit-mode
	LCD_WRITE(0x20);
	_delay_us(100);
	// 4 bit, 2 lines, 5x7 font
	LCD_CMD(0x28);
	_delay_us(100);
	// Entry mode, increment address, no shift
	LCD_CMD(0x06);
	_delay_us(100);
	// Display on, no cursor, no blink
	LCD_CMD(0x0C);
	_delay_us(100);
	// Select CG address 0x00
	LCD_CMD(0x40);
	_delay_us(50);
	// Write character patterns
	byte i;
	for (i=0; i<sizeof(LCDCharPattern); i++)
	{
		LCD_DATA(pgm_read_byte(&LCDCharPattern[i]));
		_delay_us(50);
	}
	// Clear display
	LCD_CMD(0x01);
	_delay_ms(20);
}

void LCDGotoXY(int x,int y)
{
	LCD_CMD(0x80+( (x+(y > 0 ? 64 : 0)) & 0x7F ) )
	_delay_us(50);
}

void LCDPrintChar(char Char)
{
	LCD_DATA(pgm_read_byte(&LCDCharTab[(byte)Char]));
	_delay_us(50);
}

void LCDPrintSpaces(byte Count)
{
	while (Count)
	{
		LCDPrintChar(' ');
		Count--;
	}
}

void LCDPrintString(const char *String)
{
	while (*String)
		LCDPrintChar(*String++);
}

void LCDPrintPString(const char * PROGMEM String)
{
	while (true)
	{
		char Char = (char)pgm_read_byte(String++);
		if (Char == 0)
			break;
		LCDPrintChar(Char);
	}
}

void LCDPrintWord(word Word,int Width,bool FillZero)
{
	static char String[5+1];
	WordToString(Word,String,10,Width,FillZero);
	LCDPrintString(String);
}

void LCDPrintInt(int Int,int Width)
{
	static char String[6+1];
	IntToString(Int,String,10,Width);
	LCDPrintString(String);
}

// ******************************************************************
// ****** Event Functions *******************************************
// ******************************************************************

byte WaitEvent(void)
{
	byte Event;
	while (!xQueueReceive(EventQueue,&Event,1))
		taskYIELD();
	return Event;
}

void SendEvent(byte Event)
{
	xQueueSend(EventQueue,&Event,0);
}

// ******************************************************************
// ****** The LED and Key Handler ***********************************
// ******************************************************************

static byte LEDKeyScan(void)
{
	byte Keys = 0x00;
	byte i;
	for (i=0; i<5; i++)
	{
		portENTER_CRITICAL();
		{
			if (Keys == 0x00)
			{
				Keys = PINC & 0x03;
				if (Keys != 0x03)
					Keys = ((Keys & 0x01) << 3) | i;
				else
					Keys = 0x00;
			}
			PORTC = (PORTC & 0x03) | ((4 << i) ^ 0x7C) | ((OnOffStates & (1 << i)) ? 0x80 : 0x00);
		}
		portEXIT_CRITICAL();
		vTaskDelay(1);
	}
	return Keys;
}

static void vTaskLEDKey(void *pvParameters)
{
	byte i,Key;
	while (1)
	{
		// Wait till a key is pressed
		do
		{
			LEDKeyScan();
			LEDKeyScan();
			LEDKeyScan();
			LEDKeyScan();
			Key = LEDKeyScan();
		}
		while (Key == 0);
		// Key contains the key code itself here
		// A key has been pressed
		SendEvent(Key | KEY_PRESS);
		i = 100;
		while (LEDKeyScan() != 0)
		{
			i--;
			if (i == 0)
			{
				SendEvent(Key | KEY_REPEAT);
				i = 20;
			}
		}
		// Key has been released
		SendEvent(Key | KEY_RELEASE);
	}
}

// ******************************************************************
// ****** Automatic Control *****************************************
// ******************************************************************

static void Dim(struct TDimChannel *Dimmer)
{
	if (Dimmer->CurrentPower == Dimmer->FinalPower)
		return;

	if (Dimmer->DimSpeed == 0)
	{
		Dimmer->CurrentPower = Dimmer->FinalPower;
	}
	else
	{
		Dimmer->BresAccu += Dimmer->DiffPower;
		if (Dimmer->BresAccu < Dimmer->DimSpeed)
			return;
		Dimmer->BresAccu -= Dimmer->DimSpeed;
		
		if (Dimmer->CurrentPower < Dimmer->FinalPower)
			Dimmer->CurrentPower++;
		else
			Dimmer->CurrentPower--;
	}
	Dimmer->SetPower(Dimmer->CurrentPower);
}

static void CheckDimEvents(struct TDimChannel *Dimmer,TTime LastEventCheck)
{
	// Dim events may be checked even we are in manual mode, because the dimming itself
	// is not done here
	if (Dimmer->Events->Count == 0)
		return;
	byte i = 0;
	struct TDimEvent *DimEvent1 = &Dimmer->Events->Items[i];
	for ( ; i+1<Dimmer->Events->Count; i++)
	{
		struct TDimEvent *DimEvent2 = &Dimmer->Events->Items[i+1];
		if (SystemTime >= DimEvent1->Time && SystemTime < DimEvent2->Time)
			break;
		DimEvent1 = DimEvent2;
	}
	if ((LastEventCheck == UNKNOWN_TIME) ||
		((i+1 == Dimmer->Events->Count) && (LastEventCheck < DimEvent1->Time)) ||
		((i+1 <  Dimmer->Events->Count) && (LastEventCheck < DimEvent1->Time || LastEventCheck > SystemTime)))
	{
		// Prepare dim process
		Dimmer->FinalPower = DimEvent1->Power;
		if (Dimmer->FinalPower > Dimmer->CurrentPower)
			Dimmer->DiffPower = Dimmer->FinalPower - Dimmer->CurrentPower;
		else
			Dimmer->DiffPower = Dimmer->CurrentPower - Dimmer->FinalPower;
		Dimmer->DimSpeed = 6000L*DimEvent1->DimSpeed;
		Dimmer->BresAccu = Dimmer->DiffPower >> 1;
	}
}

static void CheckSwitchEvents(struct TSwitchChannel *Switch)
{
	if (OpMode != OPMODE_AUTO)
		return;
	if (Switch->Regulated != NULL && *Switch->Regulated)
		return;
	if (Switch->Events->Count == 0)
		return;
	byte i = 0;
	struct TSwitchEvent *SwitchEvent1 = &Switch->Events->Items[i];
	for ( ; i+1<Switch->Events->Count; i++)
	{
		struct TSwitchEvent *SwitchEvent2 = &Switch->Events->Items[i+1];
		if (SystemTime >= SwitchEvent1->Time && SystemTime < SwitchEvent2->Time)
			break;
		SwitchEvent1 = SwitchEvent2;
	}
	Switch->Switch(SwitchEvent1->State);
}

static void CheckEvents(void)
{
	static TTime LastEventCheck = UNKNOWN_TIME;

	// Do nothing, if there is no valid system time
	if (SystemTime == UNKNOWN_TIME)
		return;
	// Do nothing if we still have the same time
	if (SystemTime == LastEventCheck)
		return;

	CheckDimEvents(&DimChannels[CHANNEL_DAYLIGHT],LastEventCheck);
	CheckDimEvents(&DimChannels[CHANNEL_MOONLIGHT],LastEventCheck);
	CheckSwitchEvents(&SwitchChannels[CHANNEL_CO2]);
	CheckSwitchEvents(&SwitchChannels[CHANNEL_HEATER]);
	CheckSwitchEvents(&SwitchChannels[CHANNEL_PUMP]);

	LastEventCheck = SystemTime;
}

void Regulate(void)
{
	// Regulate Temperature
	if (Config.Temp.Regulate)
	{
		if (SensorTempCal > (Config.Temp.Nominal + (Config.Temp.Hysterese >> 1)))
			HEATER_OFF
		else if (SensorTempCal < (Config.Temp.Nominal - (Config.Temp.Hysterese >> 1)))
			HEATER_ON
	}

	// Regulate pH
	if (Config.pH.Regulate)
	{
		if (SensorpHCal > (Config.pH.Nominal + (Config.pH.Hysterese >> 1)))
			// CO2 on decreases the pH value!
			CO2_ON
		else if (SensorpHCal < (Config.pH.Nominal - (Config.pH.Hysterese >> 1)))
			// CO2 off increases the pH value!
			CO2_OFF
	}
}

word CalcCalibratedValue(word ADCUncal,struct TCalPoints *CalPoints)
{
	return ((long int)CalPoints->B.Cal-(long int)CalPoints->A.Cal)*((long int)ADCUncal-(long int)CalPoints->A.Uncal) /
		   ((long int)CalPoints->B.Uncal-(long int)CalPoints->A.Uncal)+(long int)CalPoints->A.Cal;
}

word CalcTemp(long int ADC1)
{
	long int ADC2 = (ADC1*ADC1) >> 8;
	long int ADC3 = (ADC1*ADC2) >> 8;
	return (word)((636086483L-1512516L*ADC1+419030L*ADC2-62082L*ADC3) >> 16);
}

word CalcpH(long int ADC1)
{
	// Following calculation:
	// pH = 7-Usensor/58mV
	// Usensor = (Uadc-Uref)/v
	// pH = 7-(Uadc-Uref)/(v*58mV)
	// Uadc = ADC*5V/1024
	// pH = 7-(ADC*5V/1024-Uref)/(v*58mV)
	// pH = 7-(ADC*5V/1024-2.5V)/(8.13*0.058V)
	// pH = 7 - ADC*5/1024/0.4715 + 2.5/0.4715
	// pH = 7 + 2.5/0.4715 - ADC*5/1024/0.4715
	// pH*100 = 700 + 530 - ADC*1060/1024
	// pH*100 = 1230 - (ADC*1060)>>10
	// Example: ADC = 512 => pH*100 = 700
	return (word)(1230L-((ADC1*1060L)>>10));
}

static void vTaskTimer(void *pvParameters)
{
	while (true)
	{
		word ADCUncal;
		while (!xQueueReceive(ADCQueue,&ADCUncal,1))
			taskYIELD();
			
		portENTER_CRITICAL();
		{
			if ((ADCUncal & 0x8000) == 0)
			{
				SensorTempADCUncal = ADCUncal;
				SensorTempUncal = CalcTemp(SensorTempADCUncal);
				SensorTempCal = CalcTemp(CalcCalibratedValue(SensorTempADCUncal,&Config.Temp.CalPoints));
			}
			else
			{
				SensorpHADCUncal = ADCUncal & 0x7FFF;
				SensorpHUncal = CalcpH(SensorpHADCUncal);
				SensorpHCal = CalcpH(CalcCalibratedValue(SensorpHADCUncal,&Config.pH.CalPoints));
			}
			
			if (SystemTimeHundredth == 99)
			{
				SystemTimeHundredth = 0;
				// Do the regulation of temperature and pH every second
				if (OpMode == OPMODE_AUTO)
					Regulate();

				if (SystemTime != UNKNOWN_TIME)
				{
					if (SystemTimeSecond == 59)
					{
						SystemTimeSecond = 0;
						
						if (SystemTime == MAXTIME)
						{
							SystemTime = 0;
							if (SystemDate.Day != UNKNOWN_DAY)
								INC_DATE(SystemDate);
						}
						else
							SystemTime++;
						// Check events after the complete time has been updated
						CheckEvents();
					}
					else
						SystemTimeSecond++;
				}
				// Send event after the complete time has been updated
				SendEvent(EV_SECOND);
			}
			else
				SystemTimeHundredth++;

			// Dim lights after events have been evaluated
			if (OpMode == OPMODE_AUTO)
			{				
				Dim(&DimChannels[CHANNEL_DAYLIGHT]);
				Dim(&DimChannels[CHANNEL_MOONLIGHT]);
			}
		}
		portEXIT_CRITICAL();
	}
}

// The Real Time ADC Handler
void ADCHandler(void)
{
	static long TickCounter = 0;
	TickCounter += 10000000UL-CrystalCal;
	if (TickCounter < 100000000UL)
		return;
	TickCounter -= 100000000UL;

	word ADCVal = ADC;
	if ((ADMUX & (1 << MUX0)) == 0)
	{
		ADMUX |= (1 << MUX0);
	}
	else
	{
		ADCVal |= 0x8000;
		ADMUX &= ~(1 << MUX0);
	}
	ADCSRA |= (1 << ADSC);
	xQueueSendToBackFromISR(ADCQueue,&ADCVal,false);
}

// ******************************************************************
// ****** DCF77 Decoder *********************************************
// ******************************************************************

#define DCFBIT_ZERO         0
#define DCFBIT_ONE          1
#define DCFBIT_PAUSE        2
#define DCFBIT_ERROR        3
#define DCFBIT_UNKNOWN		4

#define S7_WAITPULSOFF      0
#define S7_WAITPULSON       1
#define S7_TESTPULSSTART	2
#define S7_TESTPULS			3

//  --0---1---2---3---4---5---6---7---8---9--
//   ___
//  /   \___________________________________/
//   _______
//  /       \_______________________________/

void DecodeDCF77(void)
{
	static byte SyncState = S7_WAITPULSOFF;
	static byte SyncBad;
	static word SyncTicker;
	static byte SecTicker;
	static byte DCFBit;
	
	switch (SyncState)
	{
	case S7_WAITPULSOFF:
		if (DCF77_IS_ON)
			break;
		SyncState = S7_WAITPULSON;
		break;
		
	case S7_WAITPULSON:
		if (DCF77_IS_OFF)
			break;
		SyncBad = 20;
		SyncTicker = 50;
		SecTicker = 0;
		DCFBit = DCFBIT_UNKNOWN;
		SyncState = S7_TESTPULS;
		break;
		
	case S7_TESTPULS:
		if (SyncTicker > 0)
			SyncTicker--;
		if (SyncTicker > 0)
			break;
		DBG1_OFF
		if (SecTicker == 0)
		{
			if (DCF77_IS_ON)
				DCFBit = DCFBIT_ZERO;
			else
				DCFBit = DCFBIT_PAUSE;
		}
		else if (SecTicker == 1)
		{
			if (DCF77_IS_ON)
			{
				if (DCFBit == DCFBIT_ZERO)
					DCFBit = DCFBIT_ONE;
				else
					DCFBit = DCFBIT_ERROR;
			}
		}
		else
		{
			if (DCF77_IS_ON)
				DCFBit = DCFBIT_ERROR;
		}
		if (SecTicker < 9)
		{
			SyncTicker = 100;
			SecTicker++;
			break;
		}
		SyncTicker = 50;
		SyncState = S7_TESTPULSSTART;
		if (DCFBit == DCFBIT_ERROR)
		{
			if (SyncBad < 30)
				SyncBad++;
			else
				SyncState = S7_WAITPULSOFF;
			break;
		}
		else
		{
			if (SyncBad > 0)
				SyncBad--;
			if (SyncBad > 10)
				DCFBit = DCFBIT_ERROR;
		}
		break;

	case S7_TESTPULSSTART:
		if (SyncTicker > 0)
			SyncTicker--;
		if (SyncTicker > 0)
			break;

		// This should be the active edge of the DCF77 signal
		DBG1_ON
		// Sample the edge and do a slight adjustment
		SyncTicker = DCF77_IS_ON ? 49 : 51;

		// Send the last received bit value to the asynchronous DCF77 decoder
		xQueueSendToBackFromISR(DCFQueue,&DCFBit,false);
		
		// Begin the recording of the next bit
		SecTicker = 0;
		DCFBit = DCFBIT_UNKNOWN;
		SyncState = S7_TESTPULS;
		break;
	}
}

static byte DCFBit;
static byte DCFParity;
static void DCFWaitBit(void)
{
	while (!xQueueReceive(DCFQueue,&DCFBit,1))
		taskYIELD();
	if (DCFBit == DCFBIT_ONE)
		DCFParity++;
	DEBUGDCF77CHAR('0'+DCFBit);
}

// The asynchronous DCF77 task
static void vTaskDCF(void *pvParameters)
{
	// Make variables static in order to save space on stack
	// and increase speed of execution
	static byte i;
	
	static byte DCFMinute;
	static byte DCFHour;
	static byte DCFDay;
	static byte DCFMonth;
	static byte DCFYear;
	static byte GoodDCFTime;

Error:	
	DCFMinute = 0;
	DCFHour = 0;
	DCFDay = 0;
	DCFMonth = 0;
	DCFYear = 0;
	GoodDCFTime = 0;

	DEBUGDCF77CHAR('\r');
	DEBUGDCF77CHAR('\n');

ErrorLoop:
	DCFWaitBit();
	if (DCFBit != DCFBIT_PAUSE)
		goto ErrorLoop;

	// Receive bit 0, which is always 0
	DCFWaitBit();
	if (DCFBit != DCFBIT_ZERO)
		goto ErrorLoop;
	
Loop:
	DEBUGDCF77CHAR('\r');
	DEBUGDCF77CHAR('\n');
	DEBUGDCF77CHAR('!');

	// Skip bits 1-19, which are of no interest
	DEBUGDCF77CHAR('a');
	for (i=0; i<19; i++)
	{
		DCFWaitBit();
		if (DCFBit > DCFBIT_ONE)
			goto Error;
	}
	// Receive bit 20, which is always 1
	DEBUGDCF77CHAR('b');
	DCFWaitBit();
	if (DCFBit != DCFBIT_ONE)
		goto Error;
	// Receive bits 21-28, the minute inclusive parity
	DEBUGDCF77CHAR('c');
	DCFParity = 0;
	for (i=0; i<8; i++)
	{
		DCFWaitBit();
		if (DCFBit > DCFBIT_ONE)
			goto Error;
		DCFMinute >>= 1;
		if (DCFBit)
			DCFMinute |= 0x80;
	}
	if (DCFParity & 0x01)
		goto Error;
	DCFMinute &= 0x7F;
	// Receive bits 29-35, the hour inclusive parity
	DEBUGDCF77CHAR('d');
	DCFParity = 0;
	for (i=0; i<7; i++)
	{
		DCFWaitBit();
		if (DCFBit > DCFBIT_ONE)
			goto Error;
		DCFHour >>= 1;
		if (DCFBit)
			DCFHour |= 0x40;
	}
	if (DCFParity & 0x01)
		goto Error;
	DCFHour &= 0x3F;
	// Prepare parity for complete date
	DCFParity = 0;
	// Receive bits 36-41, the day
	DEBUGDCF77CHAR('e');
	for (i=0; i<6; i++)
	{
		DCFWaitBit();
		if (DCFBit > DCFBIT_ONE)
			goto Error;
		DCFDay >>= 1;
		if (DCFBit)
			DCFDay |= 0x20;
	}
	// Skip bits 42-44, the weekday
	DEBUGDCF77CHAR('f');
	for (i=0; i<3; i++)
	{
		DCFWaitBit();
		if (DCFBit > DCFBIT_ONE)
			goto Error;
	}
	// Receive bits 45-49, the month
	DEBUGDCF77CHAR('g');
	for (i=0; i<5; i++)
	{
		DCFWaitBit();
		if (DCFBit > DCFBIT_ONE)
			goto Error;
		DCFMonth >>= 1;
		if (DCFBit)
			DCFMonth |= 0x10;
	}
	// Receive bits 50-57, the year
	DEBUGDCF77CHAR('h');
	for (i=0; i<8; i++)
	{
		DCFWaitBit();
		if (DCFBit > DCFBIT_ONE)
			goto Error;
		DCFYear >>= 1;
		if (DCFBit)
			DCFYear |= 0x80;
	}
	// Receive and check bit 58, the parity for the complete date
	DEBUGDCF77CHAR('i');
	DCFWaitBit();
	if (DCFParity & 0x01)
		goto Error;
	
	// Receive and check bit 59, which is the pause	
	DCFWaitBit();
	if (DCFBit != DCFBIT_PAUSE)
		goto Error;

	// Receive and check bit 0, which is always 0
	DCFWaitBit();
	if (DCFBit != DCFBIT_ZERO)
		goto Error;
		
	// We received a valid time and date
	DEBUGDCF77CHAR(' ');
	DEBUGDCF77BCD(DCFHour);
	DEBUGDCF77CHAR(':');
	DEBUGDCF77BCD(DCFMinute);
	DEBUGDCF77CHAR(' ');
	DEBUGDCF77BCD(DCFDay);
	DEBUGDCF77CHAR('.');
	DEBUGDCF77BCD(DCFMonth);
	DEBUGDCF77CHAR('.');
	DEBUGDCF77BCD(DCFYear);
	// Convert values from BCD to binary
	DCFMinute = BCD2BIN(DCFMinute);
	DCFHour = BCD2BIN(DCFHour);
	DCFDay = BCD2BIN(DCFDay);
	DCFMonth = BCD2BIN(DCFMonth);
	DCFYear = BCD2BIN(DCFYear);

	if (GoodDCFTime == 0)
	{
		DCFTime = TIME(DCFHour,DCFMinute);
		DCFDate.Day = DCFDay;
		DCFDate.Month = DCFMonth;
		DCFDate.Year = DCFYear;
	}
	else
	{
		if (DCFTime == MAXTIME)
		{
			DCFTime = 0;
			INC_DATE(DCFDate);
		}
		else
			DCFTime++;

		if (DCFTime != TIME(DCFHour,DCFMinute) ||
			DCFDate.Day != DCFDay ||
			DCFDate.Month != DCFMonth ||
			DCFDate.Year != DCFYear)
			goto Error;
	}
/*
	DEBUGDCF77CHAR(' ');
	DEBUGDCF77INT(0);
	DEBUGDCF77CHAR(' ');
	DEBUGDCF77INT(1);
	DEBUGDCF77CHAR(' ');
	DEBUGDCF77INT(32767);
	DEBUGDCF77CHAR(' ');
	DEBUGDCF77INT(-1);
	DEBUGDCF77CHAR(' ');
	DEBUGDCF77INT(-32768);
*/
	if (GoodDCFTime < 2)
		GoodDCFTime++;
	if (GoodDCFTime < 2)
		goto Loop;

	portENTER_CRITICAL();
	{
		DEBUGDCF77CHAR(' ');
		DEBUGDCF77BIN(SystemTimeSecond);
		DEBUGDCF77CHAR('.');
		DEBUGDCF77BIN(SystemTimeHundredth);

		static byte CalMins = 0;
		static int CalSum = 0;
		if (CalMins > 0)
		{
			// Valid range is +/- 1000ppm
			if (SystemTicks > 60000-60 && SystemTicks < 60000+60)
			{
				CalSum += SystemTicks-60000;
				//CrystalCal = ((CalMins*60000+CalSum)/(CalMins*60000)-1)*10000000;
				//CrystalCal = (CalSum/(CalMins*60000))*10000000;
				//CrystalCal = (CalSum*10000000)/(CalMins*60000);
				//CrystalCal = (CalSum*1000)/(CalMins*6);
				int Cal = (CalSum*1000L)/(CalMins*6);

				DEBUGDCF77CHAR(' ');
				DEBUGDCF77INT(CalMins);
				DEBUGDCF77CHAR(' ');
				DEBUGDCF77INT(CalSum);
				DEBUGDCF77CHAR(' ');
				DEBUGDCF77INT(Cal);

				// After 30 minutes, save the calibration value and
				// restart measurement.
				if (CalMins == 30)
				{
					CrystalCal = Cal;
					CalMins = 0;
					CalSum = 0;
				}
			}
			else
			{
				CalMins = 0;
				CalSum = 0;
			}
		}
		SystemTicks = 0;
		CalMins++;
		
		// This is the trick: Adjust the time at second = 1
		// This will ensure, that a timed event is not executed
		// twice.
		SystemTimeHundredth = 0;
		SystemTimeSecond = 1;
		SystemTime = DCFTime;
		SystemDate = DCFDate;
	}
	portEXIT_CRITICAL();
	
	goto Loop;
}

// ******************************************************************
// ****** The Real Time 1000Hz Routine ******************************
// ******************************************************************

void vApplicationTickHook(void)
{
	if (SystemTicks < 65535)
		SystemTicks++;
	ADCHandler();	
	DecodeDCF77();
}

// ******************************************************************
// ****** Handler for Manual Channel Control ************************
// ******************************************************************

#define DIM_UP			0x01
#define HAVE_DIMMED		0x02

static bool HandleSwitches(byte Event)
{
	static byte DimDayLight = 0;
	static byte DimMoonLight = 0;


	if (((Event & KEY_CODE) < SWITCHKEY_MIN) ||
	    ((Event & KEY_CODE) > SWITCHKEY_MAX))
		return false;

	if (OpMode != OPMODE_MANUAL)
	{
		if ((Event & KEY_ACTION) == KEY_PRESS)
			Beep(BEEP_WARNING);
		return true;
	}

	switch (Event)
	{
	// ------ Dim Daylight ------------------

	case KEY_DAYLIGHT | KEY_REPEAT:
		// Stop a dim process, which might be initiated
		// by a timed event
		if (!DAYLIGHT_IS_ON)
		{
			DAYLIGHT_ON
			DimChannels[CHANNEL_DAYLIGHT].CurrentPower = 1;
			SetDaylightPower(DimChannels[CHANNEL_DAYLIGHT].CurrentPower);
			DimDayLight |= DIM_UP;
		}
		if (DimDayLight & DIM_UP)
		{
			if (DimChannels[CHANNEL_DAYLIGHT].CurrentPower < 100)
			{
				DimChannels[CHANNEL_DAYLIGHT].CurrentPower++;
				SetDaylightPower(DimChannels[CHANNEL_DAYLIGHT].CurrentPower);
			}
			else
				DimDayLight &= ~DIM_UP;
		}
		else
		{
			if (DimChannels[CHANNEL_DAYLIGHT].CurrentPower > 1)
			{
				DimChannels[CHANNEL_DAYLIGHT].CurrentPower--;
				SetDaylightPower(DimChannels[CHANNEL_DAYLIGHT].CurrentPower);
			}
			else
				DimDayLight |= DIM_UP;
		}
		DimDayLight |= HAVE_DIMMED;
		break;

	case KEY_DAYLIGHT | KEY_RELEASE:
		// Stop a dim process, which might be initiated
		// by a timed event
		if ((DimDayLight & HAVE_DIMMED) == 0)
		{
			if (DAYLIGHT_IS_ON)
				DAYLIGHT_OFF
			else
				DAYLIGHT_ON
		}
		else
		{
			DimDayLight ^=  DIM_UP;
			DimDayLight &= ~HAVE_DIMMED;
		}
		break;

	// ------ Dim Moonlight -----------------
		
	case KEY_MOONLIGHT | KEY_REPEAT:
		if (!MOONLIGHT_IS_ON)
		{
			MOONLIGHT_ON
			DimChannels[CHANNEL_MOONLIGHT].CurrentPower = 1;
			SetMoonlightPower(DimChannels[CHANNEL_MOONLIGHT].CurrentPower);
			DimMoonLight |= DIM_UP;
		}
		if (DimMoonLight & DIM_UP)
		{
			if (DimChannels[CHANNEL_MOONLIGHT].CurrentPower < 100)
			{
				DimChannels[CHANNEL_MOONLIGHT].CurrentPower++;
				SetMoonlightPower(DimChannels[CHANNEL_MOONLIGHT].CurrentPower);
			}
			else
				DimMoonLight &= ~DIM_UP;
		}
		else
		{
			if (DimChannels[CHANNEL_MOONLIGHT].CurrentPower > 1)
			{
				DimChannels[CHANNEL_MOONLIGHT].CurrentPower--;
				SetMoonlightPower(DimChannels[CHANNEL_MOONLIGHT].CurrentPower);
			}
			else
				DimMoonLight |= DIM_UP;
		}
		DimMoonLight |= HAVE_DIMMED;
		break;

	case KEY_MOONLIGHT | KEY_RELEASE:
		if ((DimMoonLight & HAVE_DIMMED) == 0)
		{
			if (MOONLIGHT_IS_ON)
				MOONLIGHT_OFF
			else
				MOONLIGHT_ON
		}
		else
		{
			DimMoonLight ^=  DIM_UP;
			DimMoonLight &= ~HAVE_DIMMED;
		}
		break;

	// ------ Switch Channels ---------------			

	case KEY_CO2 | KEY_PRESS:
		if (CO2_IS_ON)
			CO2_OFF
		else
			CO2_ON
		break;

	case KEY_HEATER | KEY_PRESS:
		if (HEATER_IS_ON)
			HEATER_OFF
		else
			HEATER_ON
		break;

	case KEY_PUMP | KEY_PRESS:
		if (PUMP_IS_ON)
			PUMP_OFF
		else
			PUMP_ON
		break;
	}

	SendEvent(EV_SWITCH);
	return true;
}

// ******************************************************************
// ****** The Menu Section ******************************************
// ******************************************************************

/*
   -----------------------------------------------------------------------------
   Uhrzeit 23:59:59  ->  Uhrzeit stellen
   -----------------------------------------------------------------------------
   Datum   31.12.08  ->  Datum stellen
   -----------------------------------------------------------------------------
   Temp     25.00C  ->  Temp-Regelung   ->  Regelung  Ein/Aus
                                             Sollwert  25.00C  -> Einstellung
                                             Hysterese  1.00C  -> Einstellung
                         Alarm-Temp
                         Kalibrierung    ->  
   -----------------------------------------------------------------------------
   pH-Wert     7.00  ->  pH-Regelung     ->  Regelung  Ein/Aus
                                             Sollwert     7.00  -> Einstellung
                                             Hysterese    0.10  -> Einstellung
                         Kalibrierung    ->  Wert 1       4.00  -> Einstellung
                                             Wert 2       7.10  -> Einstellung
   -----------------------------------------------------------------------------
   Mode    Manuell       Manuell/Auto
   -----------------------------------------------------------------------------
   Licht       100%  ->  Schaltzeiten konfigurieren
   -----------------------------------------------------------------------------
   Mondlicht   100%  ->  Schaltzeiten konfigurieren
   -----------------------------------------------------------------------------
   CO2          Ein  ->  Schaltzeiten konfigurieren (falls nicht geregelt)
   -----------------------------------------------------------------------------
   Heizung      Ein  ->  Schaltzeiten konfigurieren (falls nicht geregelt)
   -----------------------------------------------------------------------------
   Pumpe        Ein  ->  Schaltzeiten konfigurieren
   -----------------------------------------------------------------------------
   DCF77         OK
   -----------------------------------------------------------------------------
   LCD-Kontrast   5      Einstellung
   -----------------------------------------------------------------------------
   Info              ->  Versionsanzeige
*/

static struct TDimChannel *DimChannel;
static byte DimEventIndex;
static struct TSwitchChannel *SwitchChannel;
static byte SwitchEventIndex;

// ****** Generic Variable Print Functions **************************

// These functions all need to have a parameter which sizeof = 2!

void PrintTime(TTime *Time)			// Width 5
{
	if (*Time <= MAXTIME)
	{
		LCDPrintWord(HOUR(*Time),2,false);
		LCDPrintChar(':');
		LCDPrintWord(MINUTE(*Time),2,true);
	}
	else
	{
		LCDPrintChar('-');
		LCDPrintChar('-');
		LCDPrintChar(':');
		LCDPrintChar('-');
		LCDPrintChar('-');
	}
}
void PrintDate(TDate *Date)			// Width 8
{
	// This might be the system date, which might be
	// changed asynchronously
	TDate LocalDate;
	portENTER_CRITICAL();
	{
		LocalDate = *Date;
	}
	portEXIT_CRITICAL();
	if (LocalDate.Day != UNKNOWN_DAY)
	{
		LCDPrintWord(LocalDate.Day,2,false);
		LCDPrintChar('.');
		LCDPrintWord(LocalDate.Month,2,true);
		LCDPrintChar('.');
		LCDPrintWord(LocalDate.Year,2,true);
	}
	else
	{
		LCDPrintChar('-');
		LCDPrintChar('-');
		LCDPrintChar('.');
		LCDPrintChar('-');
		LCDPrintChar('-');
		LCDPrintChar('.');
		LCDPrintChar('-');
		LCDPrintChar('-');
	}
}
void PrintOnOff(bool *OnOff)		// Width 3
{
	if (*OnOff)
		LCDPrintPString(PSTR("Ein"));
	else
		LCDPrintPString(PSTR("Aus"));
}
void PrintOnOffMask(word ChannelMask)	// Width 3
{
	if (OnOffStates & ChannelMask)
		LCDPrintPString(PSTR("Ein"));
	else
		LCDPrintPString(PSTR("Aus"));
}
void PrintPercentVal(byte *Val)		// Width 4
{
	LCDPrintWord(*Val,3,false);
	LCDPrintChar('%');
}
void PrintTemp(word *Temp)			// Width 7
{
	LCDPrintWord(*Temp/100,2,false);
	LCDPrintChar('.');
	LCDPrintWord(*Temp%100,2,true);
	LCDPrintChar('');
	LCDPrintChar('C');
}
void PrintTempADC(word *ADCVal)		// Width 7
{
	word Temp = CalcTemp(*ADCVal);
	PrintTemp(&Temp);
}
void PrintpH(word *pH)				// Width 5
{
	LCDPrintWord(*pH/100,2,false);
	LCDPrintChar('.');
	LCDPrintWord(*pH%100,2,true);
}
void PrintpHADC(word *ADCVal)		// Width 5
{
	word pH = CalcpH(*ADCVal);
	PrintpH(&pH);
}
void PrintLCDContrast(byte *LCDContrast)
{
	LCDPrintWord(*LCDContrast,1,false);
}
void PrintDimEvent(struct TDimEvent *DimEvent)	// Width 13
{
	PrintTime(&DimEvent->Time);
	LCDPrintChar(' ');
	LCDPrintWord(DimEvent->DimSpeed,2,false);
	LCDPrintChar(' ');
	LCDPrintWord(DimEvent->Power,3,false);
	LCDPrintChar('%');
}
void PrintSwitchEvent(struct TSwitchEvent *SwitchEvent)	// Width 13
{
	PrintTime(&SwitchEvent->Time);
	LCDPrintSpaces(5);
	PrintOnOff(&SwitchEvent->State);
}
void PrintCal(int *Cal)				// Width 7
{   // "-1000.0"
	LCDPrintInt(*Cal/10,5);
	LCDPrintChar('.');
	LCDPrintWord(abs(*Cal%10),1,true);
}

// ****** Special Menu Print Functions ******************************

void PrintSystemTime(void)
{
	TTime Time;
	byte Second;
	portENTER_CRITICAL();
	{
		Time = SystemTime;
		Second = SystemTimeSecond;
	}
	portEXIT_CRITICAL();
	if (Time <= MAXTIME)
	{
		LCDPrintWord(HOUR(Time),2,false);
		LCDPrintChar(':');
		LCDPrintWord(MINUTE(Time),2,true);
		LCDPrintChar(':');
		LCDPrintWord(Second,2,true);
	}
	else
	{
		LCDPrintChar('-');
		LCDPrintChar('-');
		LCDPrintChar(':');
		LCDPrintChar('-');
		LCDPrintChar('-');
		LCDPrintChar(':');
		LCDPrintChar('-');
		LCDPrintChar('-');
	}
}
void PrintStateDaylight(void)
{
	if (DAYLIGHT_IS_ON)
		PrintPercentVal(&DimChannels[CHANNEL_DAYLIGHT].CurrentPower);
	else
		LCDPrintPString(PSTR(" Aus"));
}
void PrintStateMoonlight(void)
{
	if (MOONLIGHT_IS_ON)
		PrintPercentVal(&DimChannels[CHANNEL_MOONLIGHT].CurrentPower);
	else
		LCDPrintPString(PSTR(" Aus"));
}
void PrintOpMode(void)
{
	switch (OpMode)
	{
	case OPMODE_AUTO:
		LCDPrintPString(PSTR("   Auto"));
		return;
	case OPMODE_MANUAL:
		LCDPrintPString(PSTR("Manuell"));
		return;
	}
	LCDPrintPString(PSTR("      ?"));
}

// ******************************************************************

void IncByte(byte *Byte,byte Min,byte Max)           { (*Byte)++;       if (*Byte < Min || *Byte > Max) *Byte = Min; }
void DecByte(byte *Byte,byte Min,byte Max)	         { (*Byte)--;       if (*Byte < Min || *Byte > Max) *Byte = Max; }
void IncWord(word *Word,word Min,word Max,word Step) { (*Word) += Step; if (*Word < Min || *Word > Max) *Word = Min; }
void DecWord(word *Word,word Min,word Max,word Step) { (*Word) -= Step; if (*Word < Min || *Word > Max) *Word = Max; }
		
void IncHour(TTime *Time)					{ if (HOUR(*Time) == 23) *Time -= TIME(23,0); else *Time += 60; }
void DecHour(TTime *Time)					{ if (HOUR(*Time) == 0)  *Time += TIME(23,0); else *Time -= 60; }
void IncMinute(TTime *Time)					{ if (MINUTE(*Time) == 59) *Time = TIME(HOUR(*Time),0);  else (*Time)++; }
void DecMinute(TTime *Time)					{ if (MINUTE(*Time) == 0)  *Time = TIME(HOUR(*Time),59); else (*Time)--; }

void *ValueToSet;
bool SetValue(int Width,const char *Message,VoidFuncPtrVoid PrintValue,VoidFuncPtrVoid IncValue,VoidFuncPtrVoid DecValue)
{
	LCDGotoXY(0,0);
	LCDPrintSpaces(16);
	LCDGotoXY(0,0);
	LCDPrintPString(Message);
	SendEvent(EV_PAINT);
	while (true)
	{
		byte Event = WaitEvent();

		if (((Event & KEY_CODE) >= SWITCHKEY_MIN) &&
		    ((Event & KEY_CODE) <= SWITCHKEY_MAX))
		{
			HandleSwitches(Event);
			continue;
		}

		switch (Event)
		{
		case EV_PAINT:
			LCDGotoXY(0,1);
			LCDPrintSpaces(16-Width);
			PrintValue();
			continue;

		case KEY_UP | KEY_PRESS:
		case KEY_UP | KEY_REPEAT:
			IncValue();
			SendEvent(EV_PAINT);
			continue;
			
		case KEY_DOWN | KEY_PRESS:
		case KEY_DOWN | KEY_REPEAT:
			DecValue();
			SendEvent(EV_PAINT);
			continue;

		case KEY_ENTER | KEY_PRESS:
			return true;

		case KEY_ESC | KEY_PRESS:
			return false;
		}
	}
}

void EditSystemTimePrint(void)		{ PrintTime((TTime*)ValueToSet); }
void EditSystemTimeIncHour(void)	{ IncHour((TTime*)ValueToSet); }
void EditSystemTimeDecHour(void)	{ DecHour((TTime*)ValueToSet); }
void EditSystemTimeIncMinute(void)	{ IncMinute((TTime*)ValueToSet); }
void EditSystemTimeDecMinute(void)	{ DecMinute((TTime*)ValueToSet); }
void EditSystemTime(void)
{
	TTime Time;
	portENTER_CRITICAL();
	{
		Time = SystemTime;
	}
	portEXIT_CRITICAL();
	if (Time == UNKNOWN_TIME)
		Time = TIME(12,0);
	ValueToSet = &Time;
	if (!SetValue(5,PSTR("Stunde:"),EditSystemTimePrint,EditSystemTimeIncHour,EditSystemTimeDecHour))
		return;
	if (!SetValue(5,PSTR("Minute:"),EditSystemTimePrint,EditSystemTimeIncMinute,EditSystemTimeDecMinute))
		return;
	portENTER_CRITICAL();
	{
		SystemTime = Time;
		SystemTimeSecond = 0;
		SystemTimeHundredth = 0;
		CheckEvents();
	}
	portEXIT_CRITICAL();
}

void EditSystemDatePrint(void)		{ PrintDate((TDate*)ValueToSet); }
void EditSystemDateIncDay(void)		{ IncByte(&((TDate*)ValueToSet)->Day,1,31); }
void EditSystemDateDecDay(void)		{ DecByte(&((TDate*)ValueToSet)->Day,1,31); }
void EditSystemDateIncMonth(void)	{ IncByte(&((TDate*)ValueToSet)->Month,1,12); }
void EditSystemDateDecMonth(void)	{ DecByte(&((TDate*)ValueToSet)->Month,1,12); }
void EditSystemDateIncYear(void)	{ IncByte(&((TDate*)ValueToSet)->Year,0,99); }
void EditSystemDateDecYear(void)	{ DecByte(&((TDate*)ValueToSet)->Year,0,99); }
void EditSystemDate(void)
{
	TDate Date;
	portENTER_CRITICAL();
	{
		Date = SystemDate;
	}
	portEXIT_CRITICAL();
	if (Date.Day == UNKNOWN_DAY)
	{
		Date.Day = 3;
		Date.Month = 10;
		Date.Year = 8;
	}
	ValueToSet = &Date;
	if (!SetValue(8,PSTR("Tag:"),EditSystemDatePrint,EditSystemDateIncDay,EditSystemDateDecDay))
		return;
	if (!SetValue(8,PSTR("Monat:"),EditSystemDatePrint,EditSystemDateIncMonth,EditSystemDateDecMonth))
		return;
	if (!SetValue(8,PSTR("Jahr:"),EditSystemDatePrint,EditSystemDateIncYear,EditSystemDateDecYear))
		return;
	portENTER_CRITICAL();
	{
		SystemDate = Date;
	}
	portEXIT_CRITICAL();
}

void EditLCDContrastPrint(void)		{ LCDPrintWord(*(byte *)ValueToSet,1,false); }
void EditLCDContrastInc(void)		{ IncByte((byte *)ValueToSet,1,9); LCD_SET_CONTRAST(*(byte *)ValueToSet); }
void EditLCDContrastDec(void)		{ DecByte((byte *)ValueToSet,1,9); LCD_SET_CONTRAST(*(byte *)ValueToSet); }
void EditLCDContrast(void)
{
	byte Contrast = Config.LCDContrast;
	ValueToSet = &Contrast;
	if (SetValue(1,PSTR("Kontrast:"),EditLCDContrastPrint,EditLCDContrastInc,EditLCDContrastDec))
		Config.LCDContrast = Contrast;
	LCD_SET_CONTRAST(Config.LCDContrast);
}

void EditDimEventPrint(void)	 	{ PrintDimEvent((struct TDimEvent*)ValueToSet); }
void EditDimEventIncHour(void)	 	{ IncHour(&((struct TDimEvent*)ValueToSet)->Time); }
void EditDimEventDecHour(void)	 	{ DecHour(&((struct TDimEvent*)ValueToSet)->Time); }
void EditDimEventIncMinute(void) 	{ IncMinute(&((struct TDimEvent*)ValueToSet)->Time); }
void EditDimEventDecMinute(void) 	{ DecMinute(&((struct TDimEvent*)ValueToSet)->Time); }
void EditDimEventIncDimSpeed(void)	{ IncByte(&((struct TDimEvent*)ValueToSet)->DimSpeed,0,99); }
void EditDimEventDecDimSpeed(void)	{ DecByte(&((struct TDimEvent*)ValueToSet)->DimSpeed,0,99); }
void EditDimEventIncPower(void)		{ IncByte(&((struct TDimEvent*)ValueToSet)->Power,0,100); }
void EditDimEventDecPower(void)		{ DecByte(&((struct TDimEvent*)ValueToSet)->Power,0,100); }
void EditDimEvent(void)
{
	// Create a local copy of the event. It is never changed from a different process.
	struct TDimEvent Event = DimChannel->Events->Items[DimEventIndex];
	ValueToSet = &Event;
	if (!SetValue(13,PSTR("Stunde:"),EditDimEventPrint,EditDimEventIncHour,EditDimEventDecHour))
		return;
	if (!SetValue(13,PSTR("Minute:"),EditDimEventPrint,EditDimEventIncMinute,EditDimEventDecMinute))
		return;
	if (!SetValue(13,PSTR("Dauer:"),EditDimEventPrint,EditDimEventIncDimSpeed,EditDimEventDecDimSpeed))
		return;
	if (!SetValue(13,PSTR("Leistung:"),EditDimEventPrint,EditDimEventIncPower,EditDimEventDecPower))
		return;
	portENTER_CRITICAL();
	{
		byte i;
		// Remove the old event from the list
		for (i=DimEventIndex+1; i<DimChannel->Events->Count; i++)
			DimChannel->Events->Items[i-1] = DimChannel->Events->Items[i];
		DimChannel->Events->Count--;
		// Search the correct position in the timeline
		for (DimEventIndex=0; DimEventIndex<DimChannel->Events->Count; DimEventIndex++)
			if (Event.Time < DimChannel->Events->Items[DimEventIndex].Time)
				break;
		// Make space for the event at the correct position
		for (i=DimChannel->Events->Count; i > DimEventIndex; i--)
			DimChannel->Events->Items[i] = DimChannel->Events->Items[i-1];
		// Insert modified event
		DimChannel->Events->Items[DimEventIndex] = Event;
		DimChannel->Events->Count++;
		// Update state of this channel
		CheckDimEvents(DimChannel,UNKNOWN_TIME);
	}
	portEXIT_CRITICAL();
	// Exit from the option menu
	SendEvent(KEY_ESC | KEY_PRESS);
	SendEvent(KEY_ESC | KEY_RELEASE);
	return;
}

void DeleteDimEvent(void)
{
	portENTER_CRITICAL();
	{
		// Remove the event from the list:
		byte i;
		for (i=DimEventIndex+1; i<DimChannel->Events->Count; i++)
			DimChannel->Events->Items[i-1] = DimChannel->Events->Items[i];
		DimChannel->Events->Count--;
		// Update state of this channel
		CheckDimEvents(DimChannel,UNKNOWN_TIME);
	}
	portEXIT_CRITICAL();
	// Exit from the option menu
	SendEvent(KEY_ESC | KEY_PRESS);
	SendEvent(KEY_ESC | KEY_RELEASE);
}

void NewDimEvent(void)
{
	// Create a new event with default values
	struct TDimEvent Event;
	Event.Time = TIME(12,0);
	Event.DimSpeed = 10;
	Event.Power = 0;
	ValueToSet = &Event;
	if (!SetValue(13,PSTR("Stunde:"),EditDimEventPrint,EditDimEventIncHour,EditDimEventDecHour))
		return;
	if (!SetValue(13,PSTR("Minute:"),EditDimEventPrint,EditDimEventIncMinute,EditDimEventDecMinute))
		return;
	if (!SetValue(13,PSTR("Dauer:"),EditDimEventPrint,EditDimEventIncDimSpeed,EditDimEventDecDimSpeed))
		return;
	if (!SetValue(13,PSTR("Leistung:"),EditDimEventPrint,EditDimEventIncPower,EditDimEventDecPower))
		return;
	portENTER_CRITICAL();
	{
		byte i;
		// Search the correct position in the timeline
		for (DimEventIndex=0; DimEventIndex<DimChannel->Events->Count; DimEventIndex++)
			if (Event.Time < DimChannel->Events->Items[DimEventIndex].Time)
				break;
		// Make space for the event at the correct position
		for (i=DimChannel->Events->Count; i > DimEventIndex+1; i--)
			DimChannel->Events->Items[i] = DimChannel->Events->Items[i-1];
		// Insert new event
		DimChannel->Events->Items[DimEventIndex] = Event;
		DimChannel->Events->Count++;
		// Update state of this channel
		CheckDimEvents(DimChannel,UNKNOWN_TIME);
	}
	portEXIT_CRITICAL();
	return;
}

void EditSwitchEventPrint(void)	 		 { PrintSwitchEvent((struct TSwitchEvent*)ValueToSet); }
void EditSwitchEventIncHour(void)	 	 { IncHour(&((struct TSwitchEvent*)ValueToSet)->Time); }
void EditSwitchEventDecHour(void)	 	 { DecHour(&((struct TSwitchEvent*)ValueToSet)->Time); }
void EditSwitchEventIncMinute(void) 	 { IncMinute(&((struct TSwitchEvent*)ValueToSet)->Time); }
void EditSwitchEventDecMinute(void) 	 { DecMinute(&((struct TSwitchEvent*)ValueToSet)->Time); }
void EditSwitchEventIncState(void)		 { IncByte(&((struct TSwitchEvent*)ValueToSet)->State,0,1); }
void EditSwitchEventDecState(void)		 { DecByte(&((struct TSwitchEvent*)ValueToSet)->State,0,1); }
void EditSwitchEvent(void)
{
	// Create a local copy of the event. It is never changed from a different process.
	struct TSwitchEvent Event = SwitchChannel->Events->Items[SwitchEventIndex];
	ValueToSet = &Event;
	if (!SetValue(13,PSTR("Stunde:"),EditSwitchEventPrint,EditSwitchEventIncHour,EditSwitchEventDecHour))
		return;
	if (!SetValue(13,PSTR("Minute:"),EditSwitchEventPrint,EditSwitchEventIncMinute,EditSwitchEventDecMinute))
		return;
	if (!SetValue(13,PSTR("Ein/Aus:"),EditSwitchEventPrint,EditSwitchEventIncState,EditSwitchEventDecState))
		return;
	portENTER_CRITICAL();
	{
		byte i;
		// Remove the old event from the list
		for (i=SwitchEventIndex+1; i<SwitchChannel->Events->Count; i++)
			SwitchChannel->Events->Items[i-1] = SwitchChannel->Events->Items[i];
		SwitchChannel->Events->Count--;
		// Search the correct position in the timeline
		for (SwitchEventIndex=0; SwitchEventIndex<SwitchChannel->Events->Count; SwitchEventIndex++)
			if (Event.Time < SwitchChannel->Events->Items[SwitchEventIndex].Time)
				break;
		// Make space for the event at the correct position
		for (i=SwitchChannel->Events->Count; i > SwitchEventIndex; i--)
			SwitchChannel->Events->Items[i] = SwitchChannel->Events->Items[i-1];
		// Insert modified event
		SwitchChannel->Events->Items[SwitchEventIndex] = Event;
		SwitchChannel->Events->Count++;
		// Update state of this channel
		CheckSwitchEvents(SwitchChannel);
	}
	portEXIT_CRITICAL();
	// Exit from the option menu
	SendEvent(KEY_ESC | KEY_PRESS);
	SendEvent(KEY_ESC | KEY_RELEASE);
	return;
}

void DeleteSwitchEvent(void)
{
	portENTER_CRITICAL();
	{
		// Remove the event from the list:
		byte i;
		for (i=SwitchEventIndex+1; i<SwitchChannel->Events->Count; i++)
			SwitchChannel->Events->Items[i-1] = SwitchChannel->Events->Items[i];
		SwitchChannel->Events->Count--;
		// Update state of this channel
		CheckSwitchEvents(SwitchChannel);
	}
	portEXIT_CRITICAL();
	// Exit from the option menu
	SendEvent(KEY_ESC | KEY_PRESS);
	SendEvent(KEY_ESC | KEY_RELEASE);
}

void NewSwitchEvent(void)
{
	// Create a new event with default values
	struct TSwitchEvent Event;
	Event.Time = TIME(12,0);
	Event.State = false;
	ValueToSet = &Event;
	if (!SetValue(13,PSTR("Stunde:"),EditSwitchEventPrint,EditSwitchEventIncHour,EditSwitchEventDecHour))
		return;
	if (!SetValue(13,PSTR("Minute:"),EditSwitchEventPrint,EditSwitchEventIncMinute,EditSwitchEventDecMinute))
		return;
	if (!SetValue(13,PSTR("Ein/Aus:"),EditSwitchEventPrint,EditSwitchEventIncState,EditSwitchEventDecState))
		return;
	portENTER_CRITICAL();
	{
		byte i;
		// Search the correct position in the timeline
		for (SwitchEventIndex=0; SwitchEventIndex<SwitchChannel->Events->Count; SwitchEventIndex++)
			if (Event.Time < SwitchChannel->Events->Items[SwitchEventIndex].Time)
				break;
		// Make space for the event at the correct position
		for (i=SwitchChannel->Events->Count; i > SwitchEventIndex+1; i--)
			SwitchChannel->Events->Items[i] = SwitchChannel->Events->Items[i-1];
		// Insert new event
		SwitchChannel->Events->Items[SwitchEventIndex] = Event;
		SwitchChannel->Events->Count++;
		// Update state of this channel
		CheckSwitchEvents(SwitchChannel);
	}
	portEXIT_CRITICAL();
	return;
}

void EditOnOffPrint(void)	{ PrintOnOff((bool *)ValueToSet); }
void EditOnOffInc(void)		{ IncByte((byte *)ValueToSet,0,1); }
void EditOnOffDec(void)		{ DecByte((byte *)ValueToSet,0,1); }
void EditTempRegulate(void)
{	
	// Create a local copy of the value to be set
	bool OnOff = Config.Temp.Regulate;
	ValueToSet = &OnOff;
	if (!SetValue(3,PSTR("Regelung:"),EditOnOffPrint,EditOnOffInc,EditOnOffDec))
		return;
	portENTER_CRITICAL();
	{
		Config.Temp.Regulate = OnOff;
		// In case of no regulation, set state defined by the events
		CheckSwitchEvents(&SwitchChannel[CHANNEL_HEATER]);
	}
	portEXIT_CRITICAL();
}
void EditpHRegulate(void)
{
	// Create a local copy of the value to be set
	bool OnOff = Config.pH.Regulate;
	ValueToSet = &OnOff;
	if (!SetValue(3,PSTR("Regelung:"),EditOnOffPrint,EditOnOffInc,EditOnOffDec))
		return;
	portENTER_CRITICAL();
	{
		Config.pH.Regulate = OnOff;
		// In case of no regulation, set state defined by the events
		CheckSwitchEvents(&SwitchChannel[CHANNEL_CO2]);
	}
	portEXIT_CRITICAL();
}

void EditTempPrint(void)	{ PrintTemp((word *)ValueToSet); }
void EditTempInc(void)		{ IncWord((word *)ValueToSet,1000,4000,10); }
void EditTempDec(void)		{ DecWord((word *)ValueToSet,1000,4000,10); }
void EditTemp(const char * PROGMEM Message,word *TempToEdit)
{
	word Temp = *TempToEdit;
	ValueToSet = &Temp;
	if (!SetValue(7,Message,EditTempPrint,EditTempInc,EditTempDec))
		return;
	*TempToEdit	= Temp;
}
void EditTempNominal(void)   { EditTemp(PSTR("Nominal Temp:"),&Config.Temp.Nominal); }
void EditTempHysterese(void) { EditTemp(PSTR("Temp Hysterese:"),&Config.Temp.Hysterese); }
void EditTempAlarmMin(void)  { EditTemp(PSTR("Alarm Temp Min:"),&Config.Temp.AlarmMin); }
void EditTempAlarmMax(void)  { EditTemp(PSTR("Alarm Temp Max:"),&Config.Temp.AlarmMax); }

void EditpHPrint(void)	{ PrintpH((word *)ValueToSet); }
void EditpHInc(void)	{ IncWord((word *)ValueToSet,0,1400,10); }
void EditpHDec(void)	{ DecWord((word *)ValueToSet,0,1400,10); }
void EditpH(const char * PROGMEM Message,word *pHToEdit)
{
	word pH = *pHToEdit;
	ValueToSet = &pH;
	if (!SetValue(5,Message,EditpHPrint,EditpHInc,EditpHDec))
		return;
	*pHToEdit	= pH;
}
void EditpHNominal(void)   { EditpH(PSTR("Nominal pH:"),&Config.pH.Nominal); }
void EditpHHysterese(void) { EditpH(PSTR("pH Hysterese:"),&Config.pH.Hysterese); }
void EditpHAlarmMin(void)  { EditpH(PSTR("Alarm pH Min:"),&Config.pH.AlarmMin); }
void EditpHAlarmMax(void)  { EditpH(PSTR("Alarm pH Max:"),&Config.pH.AlarmMax); }

void EditTempCalWordPrint(void) { PrintTempADC((word *)ValueToSet); }
void EditTempCalWordInc(void)	{ DecWord((word *)ValueToSet,0,1023,1); }
void EditTempCalWordDec(void)	{ IncWord((word *)ValueToSet,0,1023,1); }
void EditTempCalWord(const char * PROGMEM Message,word *CalWordToEdit)
{
	word CalWord = *CalWordToEdit;
	ValueToSet = &CalWord;
	if (!SetValue(7,Message,EditTempCalWordPrint,EditTempCalWordInc,EditTempCalWordDec))
		return;
	*CalWordToEdit = CalWord;
}
void EditTempCalWordACal(void)   { EditTempCalWord(PSTR("Soll-Wert A:"),&Config.Temp.CalPoints.A.Cal); }
void EditTempCalWordAUncal(void) { EditTempCalWord(PSTR("Ist-Wert A:") ,&Config.Temp.CalPoints.A.Uncal); }
void EditTempCalWordBCal(void)   { EditTempCalWord(PSTR("Soll-Wert B:"),&Config.Temp.CalPoints.B.Cal); }
void EditTempCalWordBUncal(void) { EditTempCalWord(PSTR("Ist-Wert B:") ,&Config.Temp.CalPoints.B.Uncal); }

void EditpHCalWordPrint(void) { PrintpHADC((word *)ValueToSet); }
void EditpHCalWordInc(void)	  { DecWord((word *)ValueToSet,0,1023,1); }
void EditpHCalWordDec(void)	  { IncWord((word *)ValueToSet,0,1023,1); }
void EditpHCalWord(const char * PROGMEM Message,word *CalWordToEdit)
{
	word CalWord = *CalWordToEdit;
	ValueToSet = &CalWord;
	if (!SetValue(5,Message,EditpHCalWordPrint,EditpHCalWordInc,EditpHCalWordDec))
		return;
	*CalWordToEdit = CalWord;
}
void EditpHCalWordACal(void)   { EditpHCalWord(PSTR("Soll-Wert A:"),&Config.pH.CalPoints.A.Cal); }
void EditpHCalWordAUncal(void) { EditpHCalWord(PSTR("Ist-Wert A:") ,&Config.pH.CalPoints.A.Uncal); }
void EditpHCalWordBCal(void)   { EditpHCalWord(PSTR("Soll-Wert B:"),&Config.pH.CalPoints.B.Cal); }
void EditpHCalWordBUncal(void) { EditpHCalWord(PSTR("Ist-Wert B:") ,&Config.pH.CalPoints.B.Uncal); }

void ToggleOpMode(void)
{
	if (OpMode == OPMODE_AUTO)
		OpMode = OPMODE_MANUAL;
	else
	{
		OpMode = OPMODE_AUTO;
		portENTER_CRITICAL();
		{
			CheckEvents();
		}
		portEXIT_CRITICAL();
	}
//	SendEvent(EV_PAINT);
}

// ****** Basic Menu Declarations ***********************************

struct TMenu;
struct TStaticMenu;

typedef byte (*ByteFuncPtrTMenuPtr)(struct TMenu *);
typedef void (*VoidFuncPtrTMenuPtrByte)(struct TMenu *,byte);

static void DoMenu(struct TMenu *Menu);

struct TMenu
{
	ByteFuncPtrTMenuPtr GetCount;
	VoidFuncPtrTMenuPtrByte DrawItem;
	VoidFuncPtrTMenuPtrByte DoEnter;
};

// ****** Static Menu Declarations **********************************

#define ST_NONE		0
#define ST_MENU		1
#define ST_FUNC		2

struct TStaticItem
{
	const char Name[15+1];
	VoidFuncPtrVoidPtr ValuePrintFunc;
	void *pValue;
	byte SubType;
	void *SubItem;
};
struct TStaticMenu
{
	struct TMenu Menu;
	byte ItemCount;
	struct TStaticItem *Items;
};
byte StaticMenuGetCount(struct TStaticMenu *This)
{
	byte Count = pgm_read_byte(&This->ItemCount);
	return Count;
}
void StaticMenuDrawItem(struct TStaticMenu *This,byte Index)
{
	struct TStaticItem *Items = (struct TStaticItem *)pgm_read_word(&This->Items);
	LCDPrintPString(Items[Index].Name);
	VoidFuncPtrVoidPtr ValuePrintFunc = (VoidFuncPtrVoidPtr)pgm_read_word(&Items[Index].ValuePrintFunc);
	if (ValuePrintFunc != NULL)
		ValuePrintFunc((void *)pgm_read_word(&Items[Index].pValue));
}
void StaticMenuDoEnter(struct TStaticMenu *This,byte Index)
{
	struct TStaticItem *Items = (struct TStaticItem *)pgm_read_word(&This->Items);
	byte SubType = pgm_read_byte(&Items[Index].SubType);
	if (SubType == ST_MENU)
	{
		struct TMenu *Menu = (struct TMenu *)pgm_read_word(&Items[Index].SubItem);
		if (Menu != NULL)
		{
			DoMenu(Menu);
			SendEvent(EV_PAINT);
		}
	}
	else if (SubType == ST_FUNC)
	{
		VoidFuncPtrVoid Func = (VoidFuncPtrVoid)pgm_read_word(&Items[Index].SubItem);
		if (Func != NULL)
		{
			Func();
			SendEvent(EV_PAINT);
		}
	}
}

#define BEGIN_STATIC_MENU(MenuName)							\
	struct TStaticItem MenuName##Items[] PROGMEM =			\
		{
#define END_STATIC_MENU(MenuName)							\
		};													\
	struct TStaticMenu MenuName PROGMEM =					\
		{ { (ByteFuncPtrTMenuPtr)StaticMenuGetCount,		\
			(VoidFuncPtrTMenuPtrByte)StaticMenuDrawItem,	\
			(VoidFuncPtrTMenuPtrByte)StaticMenuDoEnter }, 	\
			 sizeof(MenuName##Items)/sizeof(struct TStaticItem), MenuName##Items };
#define MENU_ITEM_SIMPLE(Text)									\
	{ Text, NULL, NULL, ST_NONE, NULL }
#define MENU_ITEM_SUBMENU(Text,SubMenu)							\
	{ Text, NULL, NULL, ST_MENU, (void *)&(SubMenu) }
#define MENU_ITEM_FUNCTION(Text,Function)						\
	{ Text, NULL, NULL, ST_FUNC, (void *)Function }
#define MENU_ITEM_VALUE(Text,PrintFunc,Value)					\
	{ Text, (VoidFuncPtrVoidPtr)PrintFunc, Value, ST_NONE, NULL }
#define MENU_ITEM_VALUE_SUBMENU(Text,PrintFunc,Value,SubMenu)	\
	{ Text, (VoidFuncPtrVoidPtr)PrintFunc, Value, ST_MENU, (void *)&(SubMenu) }
#define MENU_ITEM_VALUE_FUNCTION(Text,PrintFunc,Value,Function)	\
	{ Text, (VoidFuncPtrVoidPtr)PrintFunc, Value, ST_FUNC, (void *)(Function) }

// ****** DimEventList Menu Declarations ****************************

struct TDimEventListMenu
{
	struct TMenu Menu;
	struct TDimChannel *DimChannel;
};
byte DimEventListMenuGetCount(struct TDimEventListMenu *This)
{
	struct TDimChannel *DimChannel = (struct TDimChannel *)pgm_read_word(&This->DimChannel);
	return DimChannel->Events->Count+1;
}
void DimEventListMenuDrawItem(struct TDimEventListMenu *This,byte Index)
{
	struct TDimChannel *DimChannel = (struct TDimChannel *)pgm_read_word(&This->DimChannel);
	if (Index < DimChannel->Events->Count)
	{
		// "I 23:59 59 100%"
		LCDPrintWord(Index+1,1,false);
		LCDPrintChar(' ');
		PrintDimEvent(&DimChannel->Events->Items[Index]);
	}
	else
	{
		LCDPrintPString(PSTR("Neu"));
		LCDPrintSpaces(12);
	}
}
struct TStaticMenu DimEventOptions;
void DimEventListMenuDoEnter(struct TDimEventListMenu *This,byte Index)
{
	// Point to the objects we are dealing with:
	// 1. The list of events
	DimChannel = (struct TDimChannel *)pgm_read_word(&This->DimChannel);
	if (Index < DimChannel->Events->Count)
	{
		// 2. The event itself
		DimEventIndex = Index;
		DoMenu((struct TMenu *)&DimEventOptions);
	}
	else
		// This function does not require a valid DimEventIndex
		NewDimEvent();
	SendEvent(EV_PAINT);
}

#define DIMEVENTLIST_MENU(MenuName,DimChannel)				\
	struct TDimEventListMenu MenuName PROGMEM =				\
		{ { (ByteFuncPtrTMenuPtr)DimEventListMenuGetCount, (VoidFuncPtrTMenuPtrByte)DimEventListMenuDrawItem, (VoidFuncPtrTMenuPtrByte)DimEventListMenuDoEnter }, \
			 &DimChannel };

// ****** SwitchEventList Menu Declarations *************************

struct TSwitchEventListMenu
{
	struct TMenu Menu;
	struct TSwitchChannel *SwitchChannel;
};
byte SwitchEventListMenuGetCount(struct TSwitchEventListMenu *This)
{
	struct TSwitchChannel *SwitchChannel = (struct TSwitchChannel *)pgm_read_word(&This->SwitchChannel);
	// Hide the "New"-Entry, if the maximum count of items have been reached
	return SwitchChannel->Events->Count < MAXLISTITEMS ? SwitchChannel->Events->Count+1 : MAXLISTITEMS;
}
void SwitchEventListMenuDrawItem(struct TSwitchEventListMenu *This,byte Index)
{
	struct TSwitchChannel *SwitchChannel = (struct TSwitchChannel *)pgm_read_word(&This->SwitchChannel);
	if (Index < SwitchChannel->Events->Count)
	{
		// "I 23:59     Ein"
		LCDPrintWord(Index+1,1,false);
		LCDPrintChar(' ');
		PrintSwitchEvent(&SwitchChannel->Events->Items[Index]);
	}
	else
	{
		LCDPrintPString(PSTR("Neu"));
		LCDPrintSpaces(12);
	}
}
struct TStaticMenu SwitchEventOptions;
void SwitchEventListMenuDoEnter(struct TSwitchEventListMenu *This,byte Index)
{
	// Point to the objects we are dealing with:
	// 1. The list of events
	SwitchChannel = (struct TSwitchChannel *)pgm_read_word(&This->SwitchChannel);
	if (Index < SwitchChannel->Events->Count)
	{
		// 2. The event itself
		SwitchEventIndex = Index;
		DoMenu((struct TMenu *)&SwitchEventOptions);
	}
	else
		// This function does not require a valid SwitchEventIndex
		NewSwitchEvent();
	SendEvent(EV_PAINT);
}

#define SWITCHEVENTLIST_MENU(MenuName,SwitchChannel)			\
	struct TSwitchEventListMenu MenuName PROGMEM =				\
		{ { (ByteFuncPtrTMenuPtr)SwitchEventListMenuGetCount, (VoidFuncPtrTMenuPtrByte)SwitchEventListMenuDrawItem, (VoidFuncPtrTMenuPtrByte)SwitchEventListMenuDoEnter }, \
			 &SwitchChannel };

// ****** The Menu Structure ****************************************

// ------ Temperature Submenus --------------------------------------

BEGIN_STATIC_MENU(TempRegulateMenu)
	MENU_ITEM_VALUE_FUNCTION ("Regelung    "   ,PrintOnOff, &Config.Temp.Regulate , EditTempRegulate ),
	MENU_ITEM_VALUE_FUNCTION ("Soll    "       ,PrintTemp , &Config.Temp.Nominal  , EditTempNominal  ),
	MENU_ITEM_VALUE_FUNCTION ("Hyst.   "       ,PrintTemp , &Config.Temp.Hysterese, EditTempHysterese)
END_STATIC_MENU(TempRegulateMenu)                                  
                                                                   
BEGIN_STATIC_MENU(TempAlarmMenu)                                   
	MENU_ITEM_VALUE_FUNCTION ("Minimum "       ,PrintTemp , &Config.Temp.AlarmMin , EditTempAlarmMin ),
	MENU_ITEM_VALUE_FUNCTION ("Maximum "       ,PrintTemp , &Config.Temp.AlarmMax , EditTempAlarmMax )
END_STATIC_MENU(TempAlarmMenu)                                     
                                                                   
BEGIN_STATIC_MENU(TempCalibrateMenu)                               
	MENU_ITEM_VALUE_FUNCTION ("Soll A  "       ,PrintTempADC,&Config.Temp.CalPoints.A.Cal,   EditTempCalWordACal  ),
	MENU_ITEM_VALUE_FUNCTION ("Ist  A  "       ,PrintTempADC,&Config.Temp.CalPoints.A.Uncal, EditTempCalWordAUncal),
	MENU_ITEM_VALUE_FUNCTION ("Soll B  "       ,PrintTempADC,&Config.Temp.CalPoints.B.Cal,   EditTempCalWordBCal  ),
	MENU_ITEM_VALUE_FUNCTION ("Ist  B  "       ,PrintTempADC,&Config.Temp.CalPoints.B.Uncal, EditTempCalWordBUncal)
END_STATIC_MENU(TempCalibrateMenu)                                 
                                                                   
BEGIN_STATIC_MENU(TempMenu)                                        
	MENU_ITEM_SUBMENU  ("Temp-Regelung  ", TempRegulateMenu ),
	MENU_ITEM_SUBMENU  ("Alarm          ", TempAlarmMenu    ),
	MENU_ITEM_SUBMENU  ("Kalibrierung   ", TempCalibrateMenu)
END_STATIC_MENU(TempMenu)                                          
                                                                   
// ------ pH Submenus -----------------------------------------------

BEGIN_STATIC_MENU(pHRegulateMenu)
	MENU_ITEM_VALUE_FUNCTION ("Regelung    "   ,PrintOnOff, &Config.pH.Regulate , EditpHRegulate ),
	MENU_ITEM_VALUE_FUNCTION ("Soll      "     ,PrintpH   , &Config.pH.Nominal  , EditpHNominal  ),
	MENU_ITEM_VALUE_FUNCTION ("Hyst.     "     ,PrintpH   , &Config.pH.Hysterese, EditpHHysterese)
END_STATIC_MENU(pHRegulateMenu)                                  
                                                                   
BEGIN_STATIC_MENU(pHAlarmMenu)                                   
	MENU_ITEM_VALUE_FUNCTION ("Minimum   "     ,PrintpH   , &Config.pH.AlarmMin , EditpHAlarmMin ),
	MENU_ITEM_VALUE_FUNCTION ("Maximum   "     ,PrintpH   , &Config.pH.AlarmMax , EditpHAlarmMin )
END_STATIC_MENU(pHAlarmMenu)                                     
                                                                   
BEGIN_STATIC_MENU(pHCalibrateMenu)                               
	MENU_ITEM_VALUE_FUNCTION ("Soll A    "     ,PrintpHADC,&Config.pH.CalPoints.A.Cal,   EditpHCalWordACal  ),
	MENU_ITEM_VALUE_FUNCTION ("Ist  A    "     ,PrintpHADC,&Config.pH.CalPoints.A.Uncal, EditpHCalWordAUncal),
	MENU_ITEM_VALUE_FUNCTION ("Soll B    "     ,PrintpHADC,&Config.pH.CalPoints.B.Cal,   EditpHCalWordBCal  ),
	MENU_ITEM_VALUE_FUNCTION ("Ist  B    "     ,PrintpHADC,&Config.pH.CalPoints.B.Uncal, EditpHCalWordBUncal)
END_STATIC_MENU(pHCalibrateMenu)                                 
                                                                   
BEGIN_STATIC_MENU(pHMenu)                                        
	MENU_ITEM_SUBMENU  ("pH-Regelung    ", pHRegulateMenu ),
	MENU_ITEM_SUBMENU  ("Alarm          ", pHAlarmMenu    ),
	MENU_ITEM_SUBMENU  ("Kalibrierung   ", pHCalibrateMenu)
END_STATIC_MENU(pHMenu)                                          

// ------------------------------------------------------------------
                                                                   
BEGIN_STATIC_MENU(DimEventOptions)                                 
	MENU_ITEM_FUNCTION ("ndern         ", EditDimEvent  ),
	MENU_ITEM_FUNCTION ("Lschen        ", DeleteDimEvent)
END_STATIC_MENU(DimEventOptions)                                   
                                                                   
BEGIN_STATIC_MENU(SwitchEventOptions)                                 
	MENU_ITEM_FUNCTION ("ndern         ", EditSwitchEvent  ),
	MENU_ITEM_FUNCTION ("Lschen        ", DeleteSwitchEvent)
END_STATIC_MENU(SwitchEventOptions)                                   
                                                                   
DIMEVENTLIST_MENU(DaylightEventsMenu,DimChannels[CHANNEL_DAYLIGHT])
                                                                   
DIMEVENTLIST_MENU(MoonlightEventsMenu,DimChannels[CHANNEL_MOONLIGHT])
                                                                   
SWITCHEVENTLIST_MENU(CO2EventsMenu,SwitchChannels[CHANNEL_CO2])

SWITCHEVENTLIST_MENU(HeaterEventsMenu,SwitchChannels[CHANNEL_HEATER])

SWITCHEVENTLIST_MENU(PumpEventsMenu,SwitchChannels[CHANNEL_PUMP])

BEGIN_STATIC_MENU(InfoMenu)
	MENU_ITEM_SIMPLE         ("AquaCtrl  V"VERSION),
	MENU_ITEM_SIMPLE         ("Letzte DCF77:  "),
	MENU_ITEM_VALUE          ("Zeit      "     , PrintTime          , &DCFTime    ),
	MENU_ITEM_VALUE          ("Datum  "        , PrintDate          , &DCFDate    ),
	MENU_ITEM_VALUE          ("Quarz   "       , PrintCal           , &CrystalCal )
END_STATIC_MENU(InfoMenu)                                 

// ------ Main Menu ---------------------------------------
                                                                   
BEGIN_STATIC_MENU(MainMenu)                                        
	MENU_ITEM_VALUE_FUNCTION ("Zeit   "        , PrintSystemTime    , NULL               , EditSystemTime      ),
	MENU_ITEM_VALUE_FUNCTION ("Datum  "        , PrintDate          , &SystemDate        , EditSystemDate      ),
	MENU_ITEM_VALUE_SUBMENU  ("Temp    "       , PrintTemp          , &SensorTempCal     , TempMenu            ),
	MENU_ITEM_VALUE_SUBMENU  ("pH-Wert   "     , PrintpH            , &SensorpHCal       , pHMenu              ),
	MENU_ITEM_VALUE_FUNCTION ("Mode    "       , PrintOpMode        , NULL               , ToggleOpMode        ),
	MENU_ITEM_VALUE_SUBMENU  ("Tageslicht "    , PrintStateDaylight , NULL               , DaylightEventsMenu  ),
	MENU_ITEM_VALUE_SUBMENU  ("Mondlicht  "    , PrintStateMoonlight, NULL               , MoonlightEventsMenu ),
	MENU_ITEM_VALUE_SUBMENU  ("CO2         "   , PrintOnOffMask     , (void *)MSK_CO2    , CO2EventsMenu       ),
	MENU_ITEM_VALUE_SUBMENU  ("Heizung     "   , PrintOnOffMask     , (void *)MSK_HEATER , HeaterEventsMenu    ),
	MENU_ITEM_VALUE_SUBMENU  ("Pumpe       "   , PrintOnOffMask     , (void *)MSK_PUMP   , PumpEventsMenu      ),
	MENU_ITEM_VALUE_FUNCTION ("LCD-Kontrast  " , PrintLCDContrast   , &Config.LCDContrast, EditLCDContrast     ),
	MENU_ITEM_SUBMENU        ("Info           ", InfoMenu )
END_STATIC_MENU(MainMenu)

// ------------------------------------------------------------------

static void DoMenu(struct TMenu *Menu)
{
	byte DisplayIndex = 0;
	byte SelectIndex = 0;

	SendEvent(EV_PAINT);

	while (true)
	{
		byte Event = WaitEvent();

		if (((Event & KEY_CODE) >= SWITCHKEY_MIN) &&
		    ((Event & KEY_CODE) <= SWITCHKEY_MAX))
		{
			HandleSwitches(Event);
			continue;
		}

		switch (Event)
		{
		case EV_SECOND:
		case EV_SWITCH:
			SendEvent(EV_PAINT);
			continue;
			
		case EV_PAINT:
			{
				byte i;
				for (i=0; i<2; i++)
				{
					LCDGotoXY(0,i);
					LCDPrintChar(i == SelectIndex-DisplayIndex ? ARROW_RIGHT : ' ');
					if (DisplayIndex+i < ((ByteFuncPtrTMenuPtr)pgm_read_word(&Menu->GetCount))(Menu))
						((VoidFuncPtrTMenuPtrByte)pgm_read_word(&Menu->DrawItem))(Menu,DisplayIndex+i);
					else
						LCDPrintSpaces(15);
				}
			}
			continue;

		case KEY_UP | KEY_PRESS:
		case KEY_UP | KEY_REPEAT:
			if (SelectIndex > DisplayIndex)
			{
				SelectIndex--;
				SendEvent(EV_PAINT);
			}
			else if (DisplayIndex > 0)
			{
				DisplayIndex--;
				SelectIndex--;
				SendEvent(EV_PAINT);
			}
			continue;
			
		case KEY_DOWN | KEY_PRESS:
		case KEY_DOWN | KEY_REPEAT:
			if (SelectIndex+1 >= ((ByteFuncPtrTMenuPtr)pgm_read_word(&Menu->GetCount))(Menu))
				continue;
			if (SelectIndex > DisplayIndex)
			{
				DisplayIndex++;
				SelectIndex++;
			}
			else
				SelectIndex++;
			SendEvent(EV_PAINT);
			continue;

		case KEY_ENTER | KEY_PRESS:
			((VoidFuncPtrTMenuPtrByte)pgm_read_word(&Menu->DoEnter))(Menu,SelectIndex);
			continue;

		case KEY_ESC | KEY_PRESS:
			return;
		}
	}
}

static void vTaskUI(void *pvParameters)
{
	while (true)
		DoMenu((struct TMenu *)&MainMenu);
}

int main(void)
{
	// Port A
	// 0: AD_TEMP
	// 1: AD_PH
	// 2: LCD_E
	// 3: LCD_RS
	// 4: LCD_D4
	// 5: LCD_D5
	// 6: LCD_D6
	// 7: LCD_D7
	DDRA  = 0b11111100;
	PORTA = 0b00000000;
	
	// Port B
	// 0: SW_LIGHT
	// 1: SW_CO2
	// 2: SW_HEATER
	// 3: SW_PUMP
	// 4: BEEP
	// 5: (MOSI)
	// 6: (MISO)
	// 7: (SCK)
	DDRB  = 0b00011111;
	PORTB = 0b11100000;
	
	// Port C
	// 0: KEY_R1
	// 1: KEY_R2
	// 2: KEY_C1
	// 3: KEY_C2
	// 4: KEY_C3
	// 5: KEY_C4
	// 6: KEY_C5
	// 7: KEY_LED
	DDRC  = 0b11111100;
	PORTC = 0b01111111;
	
	// Port D
	// 0: RXD
	// 1: TXD
	// 2: DCF77
	// 3: DBG1
	// 4: LCD_V0
	// 5: -
	// 6: PWM_MOONL
	// 7: PWM_DAYL
	DDRD  = 0b11011010;
	PORTD = 0b00110111;

	// Timer 0 - Beeper
    // Setup timer 0
    // Mode 2, CTC, prescaler 1:64
    TCCR0A = (0 << COM0A1) | (0 << COM0A0) | (0 << COM0B1) | (0 << COM0B0) | (1 << WGM01) | (0 << WGM00);
    TCCR0B = (0 << FOC0A ) | (0 << FOC0B ) | (0 << WGM02 ) | (0 << CS02  ) | (1 << CS01 ) | (1 << CS00 );
    // f = 20MHz/64/(64+1) = 4807Hz
    OCR0A = 64;

	// Timer 1
	// Interrupt frequency - Multitasking scheduler, LCD contrast
	OCR1A = configCPU_CLOCK_HZ / configTICK_RATE_HZ / 8 / 2;	// 1250
	// Setup default contrast level, stage 1
	OCR1B = 700;
	// COMA off, COMB output PWM, mode 9, phase and frequency correct PWM, which counts from 0 to OCR1A, prescaler 8
    TCCR1A = (0 << COM1A1) | (0 << COM1A0) | (1 << COM1B1) | (0 << COM1B0) | (0 << WGM11) | (1 << WGM10);
    TCCR1B = (0 << ICNC1 ) | (0 << ICES1 ) | (1 << WGM13 ) | (0 << WGM12 ) | (0 << CS12 ) | (1 << CS11 ) | (0 << CS10 );
	// Enable timer interrupt - this is okay as interrupt are currently globally disabled.
	TIMSK1 |= (1<<OCIE1A);

    // Setup timer 2 - Daylight, Moonlight
    // Mode 3, fast PWM, prescaler 1:64, f = 20MHz/64/256 = 1221Hz
    TCCR2A = (1 << COM2A1) | (0 << COM2A0) | (0 << COM2B1) | (0 << COM2B0) | (1 << WGM21) | (1 << WGM20);
    TCCR2B = (0 << FOC2A ) | (0 << FOC2B ) | (0 << WGM22 ) | (1 << CS22 )  | (0 << CS21 ) | (0 << CS20 );
    OCR2A = 255;
    OCR2B = 0;

	// Setup ADC
	// Ref is AVCC, right justify, select channel 0
	ADMUX = (0 << REFS1) | (1 << REFS0) | (0 << ADLAR) | (0 << MUX4) | (0 << MUX3) | (0 << MUX2) | (0 << MUX1) | (0 << MUX0);
	// Turn on ADC, don't start, disable autotrigger, disable interrupt, prescaler /128 -> ADC clock = 156.25kHz
	ADCSRA = (1 << ADEN) | (0 << ADSC) | (0 << ADATE) | (0 << ADIE) | (1 << ADPS2) | (1 << ADPS1) | (1 << ADPS0);

	// Setup UART
    UBRR0H = (byte)(BAUDVAL0 >> 8);
    UBRR0L = (byte)(BAUDVAL0 & 0xFF);
    UCSR0C = (0 << UMSEL01) | (0 << UMSEL00) | (0 << UPM01 ) | (0 << UPM00) | (0 << USBS0) | (1 << UCSZ01) | (1 << UCSZ00) | (0 << UCPOL0);
    UCSR0A = (0 << RXC0   ) | (0 << TXC0   ) | (0 << UDRE0 ) | (0 << FE0  ) | (0 << DOR0 ) | (0 << UPE0  ) | (1 << U2X0  ) | (0 << MPCM0 );
    UCSR0B = (1 << RXCIE0 ) | (0 << TXCIE0 ) | (0 << UDRIE0) | (1 << RXEN0) | (1 << TXEN0) | (0 << UCSZ02) | (0 << RXB80 ) | (0 << TXB80 );

	LCDInit();

	InitGlobals();

	// Set default contrast level, stage 2
	LCD_SET_CONTRAST(Config.LCDContrast);
	
	xTaskCreate( vTaskLEDKey, (signed char *)"LEDKey", configMINIMAL_STACK_SIZE+ 16, NULL, tskIDLE_PRIORITY + 2, NULL );
	xTaskCreate( vTaskUI    , (signed char *)"UI"    , configMINIMAL_STACK_SIZE+128, NULL, tskIDLE_PRIORITY + 1, NULL );
	xTaskCreate( vTaskTimer , (signed char *)"Timer" , configMINIMAL_STACK_SIZE+ 16, NULL, tskIDLE_PRIORITY + 1, NULL );
	xTaskCreate( vTaskDCF   , (signed char *)"DCF"   , configMINIMAL_STACK_SIZE+ 32, NULL, tskIDLE_PRIORITY + 1, NULL );
	xTaskCreate( vTaskEEPROM, (signed char *)"EEPROM", configMINIMAL_STACK_SIZE+ 16, NULL, tskIDLE_PRIORITY + 1, NULL );
	xTaskCreate( vTaskBeep,   (signed char *)"Beeper", configMINIMAL_STACK_SIZE+ 16, NULL, tskIDLE_PRIORITY + 1, NULL );

	// Generate a startup-beep	
	Beep(BEEP_INFO);
	
	vTaskStartScheduler();
	return 0;
}
